<?php
/**
 * Gravity Form Feed Settings file.
 *
 * @file class-wpssg-feed-settings.php
 * @package   wpsyncsheets-for-gravityforms
 * @since 1.0.0
 */

// include the BrowserDetection class.
require_once WPSSG_DIR . 'lib/vendor/php-browser-detection/src/BrowserDetection.php';

GFForms::include_payment_addon_framework();

/**
 * GFFeedAddOn extended Class.
 *
 * @since 1.0.0
 */
class WPSSG_Feed_Settings extends GFFeedAddOn {
	/**
	 * WPSSG Version.
	 *
	 * @var $version
	 */
	protected $version = WPSSG_VERSION;
	/**
	 * Minimum Gravity Version.
	 *
	 * @var $min_gravityforms_version
	 */
	// phpcs:ignore
	protected $_min_gravityforms_version = '1.9';
	/**
	 * Slug.
	 *
	 * @var $slug
	 */
	// phpcs:ignore
	protected $_slug = 'ggspreadsheetaddon';
	/**
	 * Path.
	 *
	 * @var $path
	 */
	// phpcs:ignore
	protected $_path = 'wpsyncsheets-for-gravityforms/wpsyncsheets-for-gravityforms.php';
	/**
	 * Capabilities array.
	 *
	 * @var $capabilities
	 */
	// phpcs:ignore
	protected $_full_path = __FILE__;
	/**
	 * Capabilities array.
	 *
	 * @var $capabilities
	 */
	// phpcs:ignore
	protected $_title = 'WPSyncSheets For Gravity Forms';
	/**
	 * Capabilities array.
	 *
	 * @var $capabilities
	 */
	// phpcs:ignore
	protected $_short_title = 'WPSyncSheets';
	/**
	 * Sheet setting Doc.
	 *
	 * @var $doc_sheet_setting
	 */
	protected $doc_sheet_setting = 'https://docs.wpsyncsheets.com/wpssg-google-sheets-api-settings/';
	/**
	 * Ticket generate URL.
	 *
	 * @var $submit_ticket
	 */
	protected $submit_ticket = 'https://support.wpsyncsheets.com/index.php/signup?plugin=wpssg';
	/**
	 * Capabilities array.
	 *
	 * @var $capabilities
	 */
	private static $instance = null;
	/**
	 * Capabilities array.
	 *
	 * @var $capabilities
	 */
	private static $instance_api = null;
	/**
	 * Capabilities array.
	 *
	 * @var $capabilities
	 */
	protected $capabilities = array(
		'gravityforms_ggspreadsheetaddon',
		'gravityforms_ggspreadsheetaddon_uninstall',
		'gravityforms_ggspreadsheetaddon_plugin_page',
	);

	/**
	 * Permission.
	 *
	 * @var $capabilities_settings_page
	 */
	protected $capabilities_settings_page = 'gravityforms_ggspreadsheetaddon';
	/**
	 * Form Settings Capability.
	 *
	 * @var $capabilities_form_settings
	 */
	protected $capabilities_form_settings = 'gravityforms_ggspreadsheetaddon';
	/**
	 * Unistall Capability.
	 *
	 * @var $capabilities_uninstall
	 */
	protected $capabilities_uninstall = 'gravityforms_ggspreadsheetaddon_uninstall';
	/**
	 * Plugin Page Capability.
	 *
	 * @var $capabilities_plugin_page
	 */
	protected $capabilities_plugin_page = 'gravityforms_ggspreadsheetaddon_plugin_page';

	/**
	 * Create WPSSG_Feed_Settings Instance.
	 */
	public static function get_instance() {
		if ( null === self::$instance ) {
			self::$instance = new WPSSG_Feed_Settings();
		}
		return self::$instance;
	}

	/**
	 * Create WPSSG_Google_API_Functions Instance.
	 */
	public static function get_clientinstance() {
		if ( null === self::$instance_api ) {
			self::$instance_api = new WPSyncSheets_Gravity_Forms\WPSSG_Google_API_Functions();
		}
		return self::$instance_api;
	}


	/**
	 * Plugin starting point. Handles hooks, loading of language files and PayPal delayed payment support.
	 */
	public function init() {
		parent::init();
		add_action( 'gform_after_submission', array( $this, 'wpssg_after_submission' ), 10, 2 );

		add_action( 'gform_post_payment_status', array( $this, 'wpssg_admin_payment_notification' ), 10, 8 );

		add_filter( 'gravityview-inline-edit/entry-updated', array( $this, 'wpssg_frontend_entry_updated' ), 10, 3 );

		add_action( 'gform_after_update_entry', array( $this, 'wpssg_entry_updated' ), 10, 2 );

		add_action( 'wp_ajax_wpssg_sync_count_data', array( $this, 'wpssg_sync_count_data' ) );

		add_action( 'wp_ajax_wpssg_sync_data', array( $this, 'wpssg_sync_data' ) );

		add_action( 'wp_ajax_wpssg_clear_spreadsheet', array( $this, 'wpssg_clear_spreadsheet' ) );

		add_action( 'wp_ajax_wpssg_reset_settings', array( $this, 'wpssg_reset_settings' ) );

		add_action( 'wp_ajax_wpssg_getdate_preview', array( $this, 'wpssg_getdate_preview' ) );

		add_filter( 'gform_entry_list_columns', array( $this, 'wpssg_set_custom_columns' ), 10, 2 );

		add_filter( 'gform_entries_column_filter', array( $this, 'wpssg_change_column_data' ), 10, 5 );

		add_action( 'wp_ajax_wpssg_sync_single_entry_data', array( $this, 'wpssg_sync_single_entry_data' ) );

		add_filter( 'gform_entry_detail_meta_boxes', array( $this, 'wpssg_entry_page_syncbtn' ), 10, 3 );

		add_filter( 'gform_field_map_choices', array( $this, 'wpssg_mapfield_choices' ), 10, 1 );

		add_action( 'wp_ajax_wpssg_import_entries', array( $this, 'wpssg_import_entries' ) );

		add_action( 'gform_update_status', array( $this, 'wpssg_delete_entry_post' ), 10, 4 );

		add_action( 'gravityview/delete-entry/deleted', array( $this, 'wpssg_delete_view_entry' ), 10, 2 );

		add_action( 'gform_entry_list_action', array( $this, 'wpssg_entry_list_bulk_action' ), 10, 3 );

		add_action( 'gform_delete_entry', array( $this, 'wpssg_delete_view_entry' ), 10, 2 );

		add_action( 'wpssg_cron_run', array( $this, 'wpssg_auto_sync_entries' ), 10, 2 );
		add_filter( 'cron_schedules', array( $this, 'wpssg_add_cron_interval' ) );

		add_action( 'wpssg_import_cron_run', array( $this, 'wpssg_import_entries' ), 10, 2 );
		$this->get_clientinstance();
	}
	// # FEED PROCESSING -----------------------------------------------------------------------------------------------
	/**
	 * Process the feed
	 *
	 * @param array $feed The feed object to be processed.
	 * @param array $entry The entry object currently being processed.
	 * @param array $form The form object currently being processed.
	 *
	 * @return bool|void
	 */
	public function process_feed( $feed, $entry, $form ) {
		// Retrieve the name => value pairs for all fields mapped in the 'mappedFields' field map.
		$field_map = self::$instance->get_field_map_fields( $feed, 'mappedFields' );

		// Loop through the fields from the field map setting building an array of values to be passed to the third-party service.
		$merge_vars = array();
		foreach ( $field_map as $name => $field_id ) {
			// Get the field value for the specified field id.
			$merge_vars[ $name ] = self::$instance->get_field_value( $form, $entry, $field_id );
		}
	}

	/**
	 * Enqueue admin scripts.
	 *
	 * @since  1.0
	 * @access public
	 *
	 * @return array
	 */
	public function scripts() {

		$scripts = array(
			array(
				'handle'  => 'gform_googlespreadsheet_pluginsettings',
				'deps'    => array( 'jquery' ),
				'src'     => WPSSG_URL . 'assets/js/plugin_settings.js',
				'version' => self::$instance->version,
				'enqueue' => array(
					array(
						'admin_page' => array( 'plugin_settings', 'form_settings' ),
					),
				),
			),
		);
		return array_merge( parent::scripts(), $scripts );
	}

	/**
	 * Return the stylesheets which should be enqueued.
	 *
	 * @return array
	 */
	public function styles() {

		$styles = array(
			array(
				'handle'  => 'gform_googlespreadsheet_form_settings_css',
				'src'     => WPSSG_URL . 'assets/css/form_settings.css',
				'version' => self::$instance->version,
				'enqueue' => array(
					array( 'admin_page' => array( 'form_settings' ) ),
				),
			),
		);
		return array_merge( parent::styles(), $styles );
	}

	/**
	 * Configures the settings which should be rendered on the feed edit page in the Form Settings > Feed Add-On area.
	 *
	 * @return array
	 */
	public function feed_settings_fields() {

		if ( ! self::$instance_api->checkcredenatials() ) {
			$wpssg_error = self::$instance_api->getClient( 1 );
			if ( 'Invalid token format' === (string) $wpssg_error || 'invalid_grant' === (string) $wpssg_error ) {
				$wpssg_error = '<strong class="err-msg">' . esc_html__( 'Error: Invalid Token - ', 'wpssg' ) . '<a href="' . esc_url( admin_url( 'admin.php?page=wpsyncsheets-for-gravity-forms' ) ) . '">' . esc_html__( 'Click here', 'wpssg' ) . ' </a> ' . esc_html__( 'to check more information.', 'wpssg' );
			}
			return array();
		}
		$feeddata = array();
		// phpcs:disable
		if ( isset( $_GET['id'] ) ) {
			$formdata = GFAPI::get_form( sanitize_text_field( wp_unslash( $_GET['id'] ) ) );
		}
		$default_format = self::$instance_api->wpssg_option( 'date_format' ) . ' ' . self::$instance_api->wpssg_option( 'time_format' );
		if ( isset( $_GET['fid'] ) && 0 !== (int) sanitize_text_field( wp_unslash($_GET['fid']) ) ){
			$feed_id = sanitize_text_field( wp_unslash( $_GET['fid'] ) );
			$form_id = sanitize_text_field( wp_unslash( $_GET['id'] ) );
			$feeddata = GFAPI::get_feed( $feed_id, $form_id );
			if ( isset( $feeddata['meta']['submissiondate_format'] ) && ! empty( $feeddata['meta']['submissiondate_format'] ) ){
				$date_format = $feeddata['meta']['submissiondate_format'];
				if( in_array( $date_format , array( 'withtime', 'default', 'withouttime' ), true ) ){
					if( 'withouttime' === (string) $date_format )
						$default_format = self::$instance_api->wpssg_option( 'date_format' );
				}
			}
			if( isset( $feeddata['meta']['spreedsheets'] ) && isset( $feeddata['meta']['sheetselect'] ) ){
				$sheet_data = self::$instance->wpssg_check_sheet_exist( $feeddata['meta']['spreedsheets'], $feeddata['meta']['sheetselect'] );
				if ( '' === (string)$sheet_data['spreadsheetid'] || '' === (string)$sheet_data['sheetname']) {
					$feeddata['meta']['spreedsheets'] = $sheet_data['spreadsheetid'];
					$feeddata['meta']['sheetselect'] = $sheet_data['sheetname'];
					$result = self::$instance->update_feed_meta( $feed_id, $feeddata['meta'] );
				}
			}
		}
		// phpcs:enable
		$display_checkbox = false;
		foreach ( $feeddata as $fdata ) {
			if ( isset( $fdata['meta']['type'] ) && 'product' === (string) $fdata['meta']['type'] ) {
				$display_checkbox = true;
				break;
			}
		}

		$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $formdata );
		if ( $is_creditcard_enabled ) {
			$payment_array = array(
				array(
					'label'   => 'Payment Status',
					'type'    => 'checkbox',
					'name'    => 'paymentstatus',
					'tooltip' => '<h6>' . esc_html__( 'Enable', 'wpssg' ) . '</h6>' . esc_html__( 'Allow entries in google sheet when payment is paid or approved', 'wpssg' ) . '<h6>' . esc_html__( 'Disable', 'wpssg' ) . '</h6>' . esc_html__( 'Allow entries in google sheet with any payment status', 'wpssg' ),
					'choices' => array(
						array(
							'label' => 'Paid / Approved',
							'name'  => 'paymentstatus',
						),
					),
				),
				array(
					'label'   => 'Payment Status As Header',
					'type'    => 'checkbox',
					'name'    => 'paymentstatusheader',
					'tooltip' => '<h6>' . esc_html__( 'Enable', 'wpssg' ) . '</h6>' . esc_html__( 'Payment Status as header', 'wpssg' ) . '<h6>',
					'choices' => array(
						array(
							'label' => 'Enable',
							'name'  => 'paymentstatusheader',
						),
					),
				),
			);

		}

		/* Build base fields array. */
		$base_fields = array(
			'title'  => 'WPSyncSheets For Gravity Forms Feed Settings',
			'fields' => array(
				array(
					'name'          => 'feedName',
					'label'         => esc_html__( 'Feed Name', 'wpssg' ),
					'type'          => 'text',
					'required'      => true,
					'class'         => 'medium',
					'default_value' => self::$instance->get_default_feed_name(),
					'tooltip'       => '<h6>' . esc_html__( 'Name', 'wpssg' ) . '</h6>' . esc_html__( 'Enter a feed name to uniquely identify this setup.', 'wpssg' ),
				),
				array(
					'label'    => esc_html__( 'Select Spreadsheet', 'wpssg' ),
					'type'     => 'select',
					'name'     => 'spreedsheets',
					'tooltip'  => esc_html__( 'Select spreadsheet associated with your google account', 'wpssg' ),
					'class'    => 'medium',
					'choices'  => self::$instance->wpssg_list_googlespreedsheet(),
					'onchange' => "jQuery('select[name=\"_gaddon_setting_sheetselect\"]').val('');jQuery('select[name=\"_gform_setting_sheetselect\"]').val('');jQuery(this).parents('form').submit();",
				),
				array(
					'label'      => esc_html__( 'Google Spreadsheet', 'wpssg' ),
					'type'       => 'new_spreadsheet_fun',
					'name'       => 'newspreedsheets',
					'dependency' => array(
						'field'  => 'spreedsheets',
						'values' => array( 'new' ),
					),
					'tooltip'    => esc_html__( 'Create new Google spreadsheet into your google account', 'wpssg' ),
					'class'      => 'ssheet',
				),
				array(
					'label'      => esc_html__( 'Select Sheet', 'wpssg' ),
					'type'       => 'select',
					'name'       => 'sheetselect',
					'tooltip'    => esc_html__( 'Select sheet as you have created with above selected spreadsheet', 'wpssg' ),
					'class'      => 'medium',
					'dependency' => array( $this, 'wpssg_check_new_spreadsheet_value' ),
					'choices'    => self::$instance->wpssg_sheet_for_feed_setting(),
					'onchange'   => "jQuery('select[name=\"_gaddon_setting_mappedFields\"]').val('');jQuery(this).parents('form').submit();",
				),
				array(
					'label'      => 'Set Default Headers',
					'type'       => 'headers_button',
					'tooltip'    => esc_html__( 'Set default headers from form fields', 'wpssg' ),
					'dependency' => array( $this, 'wpssg_check_spreadsheet_header_value' ),
				),
				array(
					'name'       => '',
					'label'      => '',
					'type'       => 'note_ins',
					'dependency' => array( $this, 'wpssg_check_spreadsheet_header_value' ),
				),
			),
		);
		/* Build Other feed fields array. */
		$other_fields = array(
			'title'  => __( 'Other Feed Settings', 'wpssg' ),
			'class'  => 'wpssg_other_feed_settings',
			'fields' => array(
				array(
					'name'  => 'wpssg_general_settings',
					'label' => '',
					'type'  => 'wpssg_general_settings_nonce',
				),
				array(
					'label'   => 'Edit/Update',
					'type'    => 'checkbox',
					'name'    => 'editcheckbox',
					'tooltip' => esc_html__( 'Enable edit/update sheet admin option', 'wpssg' ),
					'choices' => array(
						array(
							'label' => 'Enable',
							'name'  => 'editcheckbox',
						),
					),
				),
				array(
					'name'       => 'mappedFields',
					'label'      => esc_html__( 'Map Fields', 'wpssg' ),
					'type'       => 'field_map',
					'field_map'  => self::$instance->wpssg_mapfield_for_feed_setting(),
					'merge_tags' => '',
				),
				array(
					'label'   => esc_html__( 'Submission Date', 'wpssg' ),
					'type'    => 'checkbox',
					'name'    => 'submissiondate',
					'tooltip' => esc_html__( 'Submission Date', 'wpssg' ),
					'choices' => array(
						array(
							'label'         => esc_html__( 'Enable', 'wpssg' ),
							'name'          => 'submissiondate',
							'default_value' => 0,
						),
					),
				),
				array(
					'label'   => esc_html__( 'Form ID', 'wpssg' ),
					'type'    => 'checkbox',
					'name'    => 'wpssg_form_id',
					'tooltip' => esc_html__( 'Form ID', 'wpssg' ),
					'choices' => array(
						array(
							'label'         => esc_html__( 'Enable', 'wpssg' ),
							'name'          => 'wpssg_form_id',
							'default_value' => 0,
						),
					),
				),
				array(
					'label'   => esc_html__( 'Form Title', 'wpssg' ),
					'type'    => 'checkbox',
					'name'    => 'wpssg_form_title',
					'tooltip' => esc_html__( 'Form Title', 'wpssg' ),
					'choices' => array(
						array(
							'label'         => esc_html__( 'Enable', 'wpssg' ),
							'name'          => 'wpssg_form_title',
							'default_value' => 0,
						),
					),
				),
				array(
					'label'      => esc_html__( 'Export Entries', 'wpssg' ),
					'type'       => 'start_date',
					'tooltip'    => esc_html__( 'Export existing entries as per selected date range or Export all entries by enabling checkbox.', 'wpssg' ),
					'dependency' => array( $this, 'wpssg_check_feedsetting' ),
				),
				array(
					'name'       => '',
					'label'      => esc_html__( 'Sync all existing entries', 'wpssg' ),
					'type'       => 'sync_button',
					'dependency' => array( $this, 'wpssg_check_feedsetting' ),
				),
				array(
					'label'      => esc_html__( 'Schedule Auto Sync Entries', 'wpssg' ),
					'type'       => 'schedule_auto_sync',
					'name'       => '',
					'class'      => 'medium',
					'dependency' => array( $this, 'wpssg_check_feedsetting' ),
				),
				array(
					'name'       => '',
					'label'      => esc_html__( 'Import Entries', 'wpssg' ),
					'type'       => 'import_entries',
					'dependency' => array( $this, 'wpssg_check_feedsetting' ),
				),
				array(
					'label'      => esc_html__( 'Schedule Auto Import Entries', 'wpssg' ),
					'type'       => 'schedule_auto_import',
					'name'       => '',
					'class'      => 'medium',
					'dependency' => array( $this, 'wpssg_check_feedsetting' ),
				),
				array(
					'name'       => '',
					'label'      => esc_html__( 'View & Clear Spreadsheet', 'wpssg' ),
					'type'       => 'view_clear_button',
					'dependency' => array( $this, 'wpssg_check_feed' ),
				),
				array(
					'label'   => esc_html__( 'Freeze Header', 'wpssg' ),
					'type'    => 'checkbox',
					'name'    => 'freezecheckbox',
					'tooltip' => esc_html__( 'Freeze header row (first row) of sheet', 'wpssg' ),
					'choices' => array(
						array(
							'label'         => '',
							'name'          => 'freezecheckbox',
							'default_value' => 0,
						),
					),
				),
				array(
					'label'   => esc_html__( 'Select Row Format', 'wpssg' ),
					'type'    => 'select',
					'name'    => 'row_format',
					'choices' => array(
						array(
							'label' => esc_html__( 'RAW', 'wpssg' ),
							'value' => 'row',
						),
						array(
							'label' => esc_html__( 'USER ENTERED', 'wpssg' ),
							'value' => 'userenter',
						),
					),
				),
				array(
					'label'         => esc_html__( 'Date/Time Format', 'wpssg' ),
					'type'          => 'text',
					'name'          => 'submission_date_format',
					'default_value' => $default_format,
				),
				array(
					'name'  => '',
					'label' => '',
					'type'  => 'submissiondate_format_preview',
				),
				array(
					'name'  => '',
					'label' => '',
					'type'  => 'submissiondate_format_note',
				),
			),
		);

		/*
		*Check credit cards fields within the form
		*/
		if ( $is_creditcard_enabled || $display_checkbox ) {
			$task_fields = array(
				'title'  => __( 'After Payment Settings', 'wpssg' ),
				'fields' => $payment_array,
			);
		}

		/* Build conditional logic fields array. */
		$conditional_fields = array(
			'title'  => __( 'Conditional Logic Settings', 'wpssg' ),
			'fields' => array(
				array(
					'name'           => 'feed_condition',
					'label'          => esc_html__( 'Condition', 'wpssg' ),
					'type'           => 'feed_condition',
					'checkbox_label' => esc_html__( 'Enable Condition', 'wpssg' ),
					'instructions'   => esc_html__( 'Process this simple feed if', 'wpssg' ),
				),
			),
		);

		if ( ( $is_creditcard_enabled || $display_checkbox ) && self::$instance->wpssg_check_spreadsheet_header_values() ) {
			return array( $base_fields, $other_fields, $task_fields, $conditional_fields );
		} elseif ( self::$instance->wpssg_check_spreadsheet_header_values() ) {
			return array( $base_fields, $other_fields, $conditional_fields );
		} else {
			return array( $base_fields, $conditional_fields );
		}
	}

	/**
	 * Check Create New spreadsheet value.
	 */
	public function wpssg_check_new_spreadsheet_value() {
		/* Get current feed. */
		$feed = self::$instance->get_current_feed();

		/* Get posted settings. */
		$posted_settings = self::$instance->get_posted_settings();

		if ( ! empty( $posted_settings ) ) {
			/* Show if an action is chosen */
			if ( (string) rgar( $posted_settings, 'spreedsheets' ) === 'new' || (string) rgar( $posted_settings, 'spreedsheets' ) === '' ) {
				return false;
			}
		} else {
			/* Show if an action is chosen */
			if ( (string) rgars( $feed, 'meta/spreedsheets' ) === 'new' || (string) rgars( $feed, 'meta/spreedsheets' ) === '' ) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Create New Spreadsheet into the user's google account function
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 */
	public function settings_new_spreadsheet_fun( $field, $echo = true ) {

		$html =
			'<p> ' . esc_html__( 'Spreadsheet Name', 'wpssg' ) . '</p>
			<input type="text" name="newspreadsheetlabel" value="" class="medium gaddon-setting gaddon-text" id="newspreadsheetlabel"><br/>
			<p> ' . esc_html__( 'Sheet Name', 'wpssg' ) . ' </p>
			<input type="text" name="sheetlabel" value="" class="medium gaddon-setting gaddon-text" id="sheetlabel"><br/>
			
			<input type="checkbox" class="gaddon-setting gaddon-checkbox" id="headercheckbox" name="headercheckbox" checked="checked" disabled><span class="headercheckboxspan">' . esc_html__( 'Set deafult headers as per the form fields', 'wpssg' ) . '</span><br/><br/>
			<a href="#" class="button" id="gform_google_new_spreadsheet_button">' . esc_html__( 'Create New Spreadsheet', 'wpssg' ) . '</a>';

		$html .= wp_nonce_field( 'save_general_settings', 'wpssg_general_settings' );
		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}

	/**
	 * Set instruction for the sheet headers.
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 */
	public function settings_note_ins( $field, $echo = true ) {

		$html = '<i>' . esc_html__( 'Note: Click here to set default headers as per the form fields or create manually to start with Entry Id as a first header name.', 'wpssg' ) . '</i>';
		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Set documentation link for date-time format.
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 */
	public function settings_submissiondate_format_note( $field, $echo = true ) {

		$html         = '<a href="https://wordpress.org/support/article/formatting-date-and-time/">' . esc_html__( 'Documentation on date and time formatting', 'wpssg' ) . '</a>';
		$allowed_html = wp_kses_allowed_html( 'post' );
		echo wp_kses( $html, $allowed_html );
	}
	/**
	 * Set instruction for the sheet headers.
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 */
	public function settings_submissiondate_format_preview( $field, $echo = true ) {
		// phpcs:ignore
		if( ! isset( $_GET['fid'] ) )
			return;
		// phpcs:ignore
		$feed_id        = sanitize_text_field( wp_unslash( $_GET['fid'] ) );
		// phpcs:ignore
		$form_id        = sanitize_text_field( wp_unslash( $_GET['id'] ) );
		$feeddata       = GFAPI::get_feed( $feed_id, $form_id );
		$default_format = self::$instance_api->wpssg_option( 'date_format' ) . ' ' . self::$instance_api->wpssg_option( 'time_format' );
		if ( isset( $feeddata['meta']['submissiondate_format'] ) && ! empty( $feeddata['meta']['submissiondate_format'] ) ) {
				$date_format = $feeddata['meta']['submissiondate_format'];
			if ( 'withouttime' === (string) $date_format ) {
				$default_format = self::$instance_api->wpssg_option( 'date_format' );
			}
		}
		if ( isset( $feeddata['meta']['submission_date_format'] ) && ! empty( $feeddata['meta']['submission_date_format'] ) ) {
			$default_format = $feeddata['meta']['submission_date_format'];
		}
		$date_val     = wp_date( $default_format );
		$html         = '<strong>' . esc_html__( 'Preview : ', 'wpssg' ) . '</strong><div id="wpssg_preview">' . $date_val . '</div><img src="' . esc_url( admin_url( 'images/spinner.gif' ) ) . '" id="previewloader">';
		$allowed_html = wp_kses_allowed_html( 'post' );
		echo wp_kses( $html, $allowed_html );
	}

	/**
	 * Configures the synchronization button for synchronize old entries to Google Sheet .
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_sync_button( $field, $echo = true ) {
		$html      =
			'<img src="' . esc_url( admin_url( 'images/spinner.gif' ) ) . '" id="syncloader"><span id="synctext">' . esc_html__( 'Synchronizing...', 'wpssg' ) . '</span>';
			$html .= '<a href="#" class="button" id="gform_google_sync_button">' . esc_html__( 'Click to Sync', 'wpssg' ) . '</a><br><i id="note">' . esc_html__( 'Note: Click to Sync button will be append all the existing entries to the selected sheet as above, make sure you have correctly mapped the above fields with above sheet headers.', 'wpssg' ) . '</i>';

		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Configures the Import Entries button .
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_import_entries( $field, $echo = true ) {
		$html  =
			'<img src="' . esc_url( admin_url( 'images/spinner.gif' ) ) . '" id="importloader"><span id="importtext">' . esc_html__( 'Importing...', 'wpssg' ) . '</span>';
		$html .= '<a href="#" class="button" id="gform_google_import_entries">' . esc_html__( 'Click to Import Entries', 'wpssg' ) . '</a><br><i id="importnote">' . esc_html__( 'Note: Click to Import Entries button will create entries from selected sheet.', 'wpssg' ) . '</i>';
		$html .= '<p><a href="https://docs.wpsyncsheets.com/how-to-import-entries/" target="_blank">' . esc_html__( 'How to Import Entries?', 'wpssg' ) . '</a></p>';
		$html .= wp_nonce_field( 'import_settings', 'wpssg_import_settings' );
		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Configures the View and Clear Spreadsheet button .
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_view_clear_button( $field, $echo = true ) {

		$html = sprintf(
			'<a href="#" class="button" id="gform_google_view_button" target="_blank">%1$s</a>',
			esc_html__( 'View Spreadsheet', 'wpssg' )
		);

		$html .= sprintf(
			'<a href="#" class="button" id="gform_google_clear_button">%1$s</a>',
			esc_html__( 'Clear Spreadsheet', 'wpssg' )
		);
		$html .=
			'<img src="' . esc_url( admin_url( 'images/spinner.gif' ) ) . '" id="clearloader"><span id="cleartext">' . esc_html__( 'Clearing...', 'wpssg' ) . '</span>';
		$html .= wp_nonce_field( 'save_general_settings', 'wpssg_general_settings' );
		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Configures the settings nonce .
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_wpssg_general_settings_nonce( $field, $echo = true ) {

		$html = wp_nonce_field( 'save_general_settings', 'wpssg_general_settings' );
		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Configures the synchronization button for synchronize old entries to Google Sheet .
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_headers_button( $field, $echo = true ) {

		$html  = sprintf(
			' <a href="#" class="button" id="gform_google_header_button">%1$s</a><br>',
			esc_html__( 'Click Here', 'wpssg' )
		);
		$html .= wp_nonce_field( 'save_general_settings', 'wpssg_general_settings' );
		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}

	/**
	 * Configures the date field .
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_start_date( $field, $echo = true ) {

		$html = "<div class='gaddon-setting-checkbox'><input type='checkbox' id='syncallentries' name='syncallentries' checked='checked'><label for='syncallentries'><span>" . esc_html__( 'All Entries', 'wpssg' ) . "</span></label><br><input type='date' class='wpssgdaterange' name='start_date' placeholder='" . esc_html__( 'Start Date', 'wpssg' ) . "' ><input type='date' name='end_date' placeholder='" . esc_html__( 'End Date', 'wpssg' ) . "' class='wpssgdaterange'></div>";

		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Configures the Schedule Auto Sync Entries.
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_schedule_auto_sync( $field, $echo = true ) {

		$settings = array();
		// @codingStandardsIgnoreStart.
		if ( isset( $_GET['id'] ) && isset( $_GET['fid'] ) && 0 !== (int) sanitize_text_field( wp_unslash( $_GET['fid'] ) ) ) {
			$wpssg_fid = sanitize_text_field( wp_unslash( $_GET['fid'] ) );
			$form_id   = sanitize_text_field( wp_unslash( $_GET['id'] ) );
			// @codingStandardsIgnoreEnd.
			$feeddata = GFAPI::get_feed( $wpssg_fid, $form_id );

			$settings = $feeddata['meta'] ? $feeddata['meta'] : array();
		}
		$wpssg_scheduling_enable            = isset( $settings ['wpssg_scheduling_enable'] ) ? $settings ['wpssg_scheduling_enable'] : '';
		$wpssg_scheduling_run_on            = isset( $settings ['wpssg_scheduling_run_on'] ) ? $settings ['wpssg_scheduling_run_on'] : '';
		$wpssg_schedule_recurrence          = isset( $settings ['wpssg_schedule_recurrence'] ) ? $settings ['wpssg_schedule_recurrence'] : '';
		$wpssg_scheduling_weekly_days       = isset( $settings ['wpssg_scheduling_weekly_days'] ) ? $settings ['wpssg_scheduling_weekly_days'] : '';
		$wpssg_scheduling_date              = isset( $settings ['wpssg_scheduling_date'] ) ? $settings ['wpssg_scheduling_date'] : '';
		$wpssg_scheduling_time              = isset( $settings ['wpssg_scheduling_time'] ) ? $settings ['wpssg_scheduling_time'] : '';
		$wpssg_scheduling_weekly_days_array = array();
		if ( ! empty( $wpssg_scheduling_weekly_days ) ) {
			$wpssg_scheduling_weekly_days_array = explode( ',', $wpssg_scheduling_weekly_days );
		}
		$wpssg_active     = '';
		$wpssg_deactive   = '';
		$wpssg_recurrence = '';
		$wpssg_weekly     = '';
		$wpssg_onetime    = '';
		if ( 1 === (int) $wpssg_scheduling_enable ) {
			$wpssg_active = 'checked=checked';
		} else {
			$wpssg_deactive = 'checked=checked';
		}

		if ( 'recurrence' === (string) $wpssg_scheduling_run_on ) {
			$wpssg_recurrence = 'checked=checked';
		}
		if ( 'onetime' === (string) $wpssg_scheduling_run_on ) {
			$wpssg_onetime = 'checked=checked';
		}
		if ( 'weekly' === (string) $wpssg_scheduling_run_on ) {
			$wpssg_weekly = 'checked=checked';
		}
		if ( empty( $wpssg_recurrence ) && empty( $wpssg_onetime ) && empty( $wpssg_weekly ) ) {
			$wpssg_recurrence = 'checked=checked';
		}
		$wpssg_interval = array(
			'thirty_minutes' => esc_html__( 'Every Thirty Minutes', 'wpssg' ),
			'once_daily'     => esc_html__( 'Once Daily', 'wpssg' ),
			'twice_daily'    => esc_html__( 'Twice Daily', 'wpssg' ),
			'fifteen_days'   => esc_html__( 'Every Fifteen Days', 'wpssg' ),
			'monthly'        => esc_html__( 'Monthly', 'wpssg' ),
		);
		$wpssg_weekday  = array(
			'monday'    => esc_html__( 'Mon', 'wpssg' ),
			'tuesday'   => esc_html__( 'Tue', 'wpssg' ),
			'wednesday' => esc_html__( 'Wed', 'wpssg' ),
			'thursday'  => esc_html__( 'Thu', 'wpssg' ),
			'friday'    => esc_html__( 'Fri', 'wpssg' ),
			'saturday'  => esc_html__( 'Sat', 'wpssg' ),
			'sunday'    => esc_html__( 'Sun', 'wpssg' ),
		);

		$html = '<div class="schedulecontainer">
					<input type="radio" name="scheduling_enable" value="0" ' . esc_attr( $wpssg_deactive ) . ' id="donotschedule">
					<label for="donotschedule">' . esc_html__( 'Do Not Schedule', 'wpssg' ) . '</label>
					<input type="radio" name="scheduling_enable" value="1" id="autoschedule" ' . esc_attr( $wpssg_active ) . '>
					<label for="autoschedule">' . esc_html__( 'Automatic Scheduling', 'wpssg' ) . '</label>
				</div>
				<div id="automatic-scheduling">
					<div class="schedulecontainer">
						<div class="input">
							<input type="radio" name="scheduling_run_on" value="recurrence" id="recurrenceradio" ' . esc_attr( $wpssg_recurrence ) . ' >
							<label for="recurrenceradio">' . esc_html__( 'Schedule Recurrence', 'wpssg' ) . '</label>
						</div>
						<select name="schedule_recurrence" id="schedule_recurrence">';
		foreach ( $wpssg_interval as $intkey => $intvalue ) {
			$selected = '';
			if ( (string) $intkey === (string) $wpssg_schedule_recurrence ) {
				$selected = 'selected="selected"';
			}
			$html .= '
								<option value="' . esc_attr( $intkey ) . '" ' . esc_attr( $selected ) . '>' . esc_html( $intvalue ) . '</option>';
		}
						$html .= '</select>
					</div>
					<div class="schedulecontainer">
						<div class="input">
							<input type="radio" name="scheduling_run_on" value="weekly" id="weeklyradio" ' . esc_attr( $wpssg_weekly ) . '>
							<label for="weeklyradio">' . esc_html__( 'Every week on...', 'wpssg' ) . '</label>
						</div>
						<input type="hidden" name="scheduling_weekly_days" value="' . esc_attr( $wpssg_scheduling_weekly_days ) . '" id="weekly_days">
						<ul class="days-of-week" id="weekly">';

		foreach ( $wpssg_weekday as $weekname => $weekshortname ) {
			$selectedday = '';
			if ( in_array( $weekname, $wpssg_scheduling_weekly_days_array, true ) ) {
				$selectedday = 'selected';
			}

			$html .= '<li data-day="' . esc_attr( $weekname ) . '" class="' . esc_attr( $selectedday ) . ' button">' . esc_html( $weekshortname ) . '</li>';
		}
						$html .= '</ul>
					</div>
					<div class="schedulecontainer">
						<div class="input">
							<input type="radio" name="scheduling_run_on" value="onetime" id="onetime" ' . esc_attr( $wpssg_onetime ) . '>
							<label for="onetime">' . esc_html__( 'One time run at', 'wpssg' ) . '</label>
						</div>
						<div id="scheduling_date">
							<input type="date" name="scheduling_date" autocomplete="off" value="' . esc_attr( $wpssg_scheduling_date ) . '">
							<input type="time" name="scheduling_time" autocomplete="off" value="' . esc_attr( $wpssg_scheduling_time ) . '">
						</div>
					</div>
				</div>';

		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}

	/**
	 * Configures the Schedule Auto Import Entries.
	 *
	 * @param string $field Field.
	 * @param bool   $echo Echo or not.
	 * @return array
	 */
	public function settings_schedule_auto_import( $field, $echo = true ) {

		$settings = array();
		// @codingStandardsIgnoreStart.
		if ( isset( $_GET['id'] ) && isset( $_GET['fid'] ) && 0 !== (int) sanitize_text_field( wp_unslash( $_GET['fid'] ) ) ) {
			$wpssg_fid = sanitize_text_field( wp_unslash( $_GET['fid'] ) );
			$form_id   = sanitize_text_field( wp_unslash( $_GET['id'] ) );
			// @codingStandardsIgnoreEnd.
			$feeddata = GFAPI::get_feed( $wpssg_fid, $form_id );

			$settings = $feeddata['meta'] ? $feeddata['meta'] : array();
		}
		$wpssg_import_scheduling_enable            = isset( $settings ['wpssg_import_scheduling_enable'] ) ? $settings ['wpssg_import_scheduling_enable'] : '';
		$wpssg_import_scheduling_run_on            = isset( $settings ['wpssg_import_scheduling_run_on'] ) ? $settings ['wpssg_import_scheduling_run_on'] : '';
		$wpssg_import_schedule_recurrence          = isset( $settings ['wpssg_import_schedule_recurrence'] ) ? $settings ['wpssg_import_schedule_recurrence'] : '';
		$wpssg_import_scheduling_weekly_days       = isset( $settings ['wpssg_import_scheduling_weekly_days'] ) ? $settings ['wpssg_import_scheduling_weekly_days'] : '';
		$wpssg_import_scheduling_date              = isset( $settings ['wpssg_import_scheduling_date'] ) ? $settings ['wpssg_import_scheduling_date'] : '';
		$wpssg_import_scheduling_time              = isset( $settings ['wpssg_import_scheduling_time'] ) ? $settings ['wpssg_import_scheduling_time'] : '';
		$wpssg_import_scheduling_weekly_days_array = array();
		if ( ! empty( $wpssg_import_scheduling_weekly_days ) ) {
			$wpssg_import_scheduling_weekly_days_array = explode( ',', $wpssg_import_scheduling_weekly_days );
		}
		$wpssg_active     = '';
		$wpssg_deactive   = '';
		$wpssg_recurrence = '';
		$wpssg_weekly     = '';
		$wpssg_onetime    = '';
		if ( 1 === (int) $wpssg_import_scheduling_enable ) {
			$wpssg_active = 'checked=checked';
		} else {
			$wpssg_deactive = 'checked=checked';
		}

		if ( 'recurrence' === (string) $wpssg_import_scheduling_run_on ) {
			$wpssg_recurrence = 'checked=checked';
		}
		if ( 'onetime' === (string) $wpssg_import_scheduling_run_on ) {
			$wpssg_onetime = 'checked=checked';
		}
		if ( 'weekly' === (string) $wpssg_import_scheduling_run_on ) {
			$wpssg_weekly = 'checked=checked';
		}
		if ( empty( $wpssg_recurrence ) && empty( $wpssg_onetime ) && empty( $wpssg_weekly ) ) {
			$wpssg_recurrence = 'checked=checked';
		}
		$wpssg_interval = array(
			'thirty_minutes' => esc_html__( 'Every Thirty Minutes', 'wpssg' ),
			'once_daily'     => esc_html__( 'Once Daily', 'wpssg' ),
			'twice_daily'    => esc_html__( 'Twice Daily', 'wpssg' ),
			'fifteen_days'   => esc_html__( 'Every Fifteen Days', 'wpssg' ),
			'monthly'        => esc_html__( 'Monthly', 'wpssg' ),
		);
		$wpssg_weekday  = array(
			'monday'    => esc_html__( 'Mon', 'wpssg' ),
			'tuesday'   => esc_html__( 'Tue', 'wpssg' ),
			'wednesday' => esc_html__( 'Wed', 'wpssg' ),
			'thursday'  => esc_html__( 'Thu', 'wpssg' ),
			'friday'    => esc_html__( 'Fri', 'wpssg' ),
			'saturday'  => esc_html__( 'Sat', 'wpssg' ),
			'sunday'    => esc_html__( 'Sun', 'wpssg' ),
		);

		$html = '<div class="import_schedulecontainer">
					<input type="radio" name="import_scheduling_enable" value="0" ' . esc_attr( $wpssg_deactive ) . ' id="import_donotschedule">
					<label for="import_donotschedule">' . esc_html__( 'Do Not Schedule', 'wpssg' ) . '</label>
					<input type="radio" name="import_scheduling_enable" value="1" id="import_autoschedule" ' . esc_attr( $wpssg_active ) . '>
					<label for="import_autoschedule">' . esc_html__( 'Automatic Scheduling', 'wpssg' ) . '</label>
				</div>
				<div id="import-automatic-scheduling">
					<div class="import_schedulecontainer">
						<div class="input">
							<input type="radio" name="import_scheduling_run_on" value="recurrence" id="import_recurrenceradio" ' . esc_attr( $wpssg_recurrence ) . ' >
							<label for="import_recurrenceradio">' . esc_html__( 'Schedule Recurrence', 'wpssg' ) . '</label>
						</div>
						<select name="import_schedule_recurrence" id="import_schedule_recurrence">';
		foreach ( $wpssg_interval as $intkey => $intvalue ) {
			$selected = '';
			if ( (string) $intkey === (string) $wpssg_import_schedule_recurrence ) {
				$selected = 'selected="selected"';
			}
			$html .= '
								<option value="' . esc_attr( $intkey ) . '" ' . esc_attr( $selected ) . '>' . esc_html( $intvalue ) . '</option>';
		}
						$html .= '</select>
					</div>
					<div class="import_schedulecontainer">
						<div class="input">
							<input type="radio" name="import_scheduling_run_on" value="weekly" id="import_weeklyradio" ' . esc_attr( $wpssg_weekly ) . '>
							<label for="import_weeklyradio">' . esc_html__( 'Every week on...', 'wpssg' ) . '</label>
						</div>
						<input type="hidden" name="import_scheduling_weekly_days" value="' . esc_attr( $wpssg_import_scheduling_weekly_days ) . '" id="import_weekly_days">
						<ul class="import-days-of-week" id="import_weekly">';

		foreach ( $wpssg_weekday as $weekname => $weekshortname ) {
			$selectedday = '';
			if ( in_array( $weekname, $wpssg_import_scheduling_weekly_days_array, true ) ) {
				$selectedday = 'selected';
			}

			$html .= '<li data-day="' . esc_attr( $weekname ) . '" class="' . esc_attr( $selectedday ) . ' button">' . esc_html( $weekshortname ) . '</li>';
		}
						$html .= '</ul>
					</div>
					<div class="import_schedulecontainer">
						<div class="input">
							<input type="radio" name="import_scheduling_run_on" value="onetime" id="import_onetime" ' . esc_attr( $wpssg_onetime ) . '>
							<label for="import_onetime">' . esc_html__( 'One time run at', 'wpssg' ) . '</label>
						</div>
						<div id="import_scheduling_date">
							<input type="date" name="import_scheduling_date" autocomplete="off" value="' . esc_attr( $wpssg_import_scheduling_date ) . '">
							<input type="time" name="import_scheduling_time" autocomplete="off" value="' . esc_attr( $wpssg_import_scheduling_time ) . '">
						</div>
					</div>
				</div>';

		if ( $echo ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $html, $allowed_html );
		}
		return $html;
	}
	/**
	 * Configures which columns should be displayed on the feed list page.
	 *
	 * @return array
	 */
	public function feed_list_columns() {
		return array(
			'feedName' => esc_html__( 'Name', 'wpssg' ),
		);
	}

	/**
	 * Prevent feeds being listed or created if an authToken isn't blank.
	 *
	 * @return bool
	 */
	public function can_create_feed() {
		// Get the plugin settings.
		if ( ! self::$instance_api->getClient() ) {
			return false;
		}
		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );
		// Access a specific setting e.g. an authToken.
		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) ) {
			return true;
		} else {
			return false;
		}
	}
	/**
	 * Configure Addon Message.
	 */
	public function configure_addon_message() {
		$settings_link = '<a href="' . esc_url( admin_url( 'admin.php?page=wpsyncsheets-for-gravity-forms' ) ) . '">' . esc_html__( 'WPSynSheets Google API Settings', 'wpssg' ) . '</a>.';
		return esc_html__( 'To get started, please configure your ', 'wpssg' ) . $settings_link;
	}


	/**
	 * Count the number of entries to sync.
	 */
	public function wpssg_sync_count_data() {
		if ( ! isset( $_POST['_wpnonce'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['_wpnonce'] ) ), 'save_general_settings' ) ) {
			echo esc_html( 'error' );
			wp_die();
		}

		if ( isset( $_POST['fid'] ) && ! empty( $_POST['fid'] ) ) {
			$wpssg_fid = (int) sanitize_text_field( wp_unslash( $_POST['fid'] ) );
		}
		if ( isset( $_POST['form_id'] ) && ! empty( $_POST['form_id'] ) ) {
			$form_id = sanitize_text_field( wp_unslash( $_POST['form_id'] ) );
		}
		$syncformfeed    = RGFormsModel::get_form_meta( $form_id );
		$formfeed        = self::$instance->get_feeds( $form_id );
		$form_field_type = self::$instance->wpssg_form_field_type( $syncformfeed );
		$wpssg_form_data = GFAPI::get_form( $form_id );

		foreach ( $formfeed as $feedsetting ) {
			if ( (int) $feedsetting['id'] === (int) $wpssg_fid ) {
				$settings = $feedsetting['meta'];
			}
		}
		$paging          = array(
			'offset'    => 0,
			'page_size' => 999999,
		);
		$search_criteria = array( 'status' => 'active' );
		if ( isset( $_POST['syncallentries'] ) ) {
			if ( 'false' === sanitize_text_field( wp_unslash( $_POST['syncallentries'] ) ) && ! empty( sanitize_text_field( wp_unslash( $_POST['start_date'] ) ) ) && ! empty( sanitize_text_field( wp_unslash( $_POST['end_date'] ) ) ) ) {
				$search_criteria['start_date'] = sanitize_text_field( wp_unslash( $_POST['start_date'] ) );
				$search_criteria['end_date']   = sanitize_text_field( wp_unslash( $_POST['end_date'] ) );
			}
		}

		// default.
		$sorting = array( 'direction' => 'ASC' );

		$entry = GFAPI::get_entries( $form_id, $search_criteria, $sorting, $paging );

		$total_entries = GFAPI::count_entries( $form_id, $search_criteria );

		$entryids            = array_column( $entry, 'id' );
		$count_total_entries = count( $entryids );

		if ( isset( $_POST['spreedsheets'] ) && ! empty( $_POST['spreedsheets'] ) ) {
			$spreadsheetid = sanitize_text_field( wp_unslash( $_POST['spreedsheets'] ) );
		}

		if ( isset( $_POST['sheetselect'] ) && ! empty( $_POST['sheetselect'] ) ) {
			$range = sanitize_text_field( wp_unslash( $_POST['sheetselect'] ) );
		}
		$allentry              = self::$instance_api->get_row_list( $spreadsheetid, $range . '!A:Z' );
		$data                  = $allentry->getValues();
		$data                  = array_map(
			function( $element ) {
				if ( isset( $element['0'] ) ) {
					return $element['0'];
				} else {
					return '';
				}
			},
			$data
		);
		$entrycount            = 0;
		$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );

		foreach ( $entryids as $entryid ) {
			$is_condition_enabled = rgar( $settings, 'feed_condition_conditional_logic' ) === '1';
			$logic                = rgars( $settings, 'feed_condition_conditional_logic_object/conditionalLogic' );
			$entry                = GFAPI::get_entry( $entryid );
			if ( $is_condition_enabled && ! empty( $logic ) ) {
				$check = GFCommon::evaluate_conditional_logic( $logic, $wpssg_form_data, $entry );
				if ( ! $check ) {
					continue;
				}
			}
			if ( $is_creditcard_enabled ) {
				if ( isset( $settings['paymentstatus'] ) && $settings['paymentstatus'] ) {
					if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
						continue;
					}
				}
			}
			$data = self::$instance->wpssg_convert_int( $data );

			if ( in_array( (int) $entryid, $data, true ) ) {
				continue;
			} else {
				$entrycount++;
			}
		}
		$requset = array();
		$requset = array(
			'totalentries'    => $entrycount,
			'sheetname'       => $range,
			'form_id'         => $form_id,
			'search_criteria' => $search_criteria,
		);
		echo wp_json_encode( $requset );
		die;
	}
	/**
	 * Sync entries to the sheet.
	 */
	public function wpssg_sync_data() {
		if ( ! isset( $_POST['sync_nonce_token'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['sync_nonce_token'] ) ), 'sync_nonce' )
		) {
			wp_die();
		}

		if ( isset( $_POST['form_id'] ) ) {
			$form_id = (int) sanitize_text_field( wp_unslash( $_POST['form_id'] ) );
		}
		if ( isset( $_POST['fid'] ) ) {
			$wpssg_fid = (int) sanitize_text_field( wp_unslash( $_POST['fid'] ) );
		}
		$syncformfeed    = RGFormsModel::get_form_meta( $form_id );
		$wpssg_form_data = GFAPI::get_form( $form_id );
		$formfeed        = self::$instance->get_feeds( $form_id );
		$form_field_type = self::$instance->wpssg_form_field_type( $syncformfeed );

		foreach ( $formfeed as $feedsetting ) {
			if ( (int) $feedsetting['id'] === (int) $wpssg_fid ) {
				$settings = $feedsetting['meta'];
			}
		}

		$paging = array(
			'offset'    => 0,
			'page_size' => 999999,
		);

		if ( isset( $_POST['search_criteria'] ) ) {
			$search_criteria = (array) json_decode( stripslashes( sanitize_text_field( wp_unslash( $_POST['search_criteria'] ) ) ) );
		}

		// default.
		$sorting = array( 'direction' => 'ASC' );

		$entry = GFAPI::get_entries( $form_id, $search_criteria, $sorting, $paging );

		if ( isset( $_POST['entry_limit'] ) ) {
			$entry_limit = sanitize_text_field( wp_unslash( $_POST['entry_limit'] ) );
		}
		$entryids = array_column( $entry, 'id' );

		$spreadsheetid         = $settings['spreedsheets'];
		$range                 = $settings['sheetselect'];
		$formmapfieldsval      = self::$instance->wpssg_get_mapfields( $spreadsheetid, $range );
		$allentry              = self::$instance_api->get_row_list( $spreadsheetid, $range . '!A:Z' );
		$data                  = $allentry->getValues();
		$data                  = array_map(
			function( $element ) {
				if ( isset( $element['0'] ) ) {
					return $element['0'];
				} else {
					return '';
				}
			},
			$data
		);
		$data                  = self::$instance->wpssg_convert_int( $data );
		$newenrty              = array();
		$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );
		$val                   = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $settings );
		$is_condition_enabled  = rgar( $settings, 'feed_condition_conditional_logic' ) === '1';
		$logic                 = rgars( $settings, 'feed_condition_conditional_logic_object/conditionalLogic' );

		foreach ( $entryids as $entryid ) {
			if ( in_array( (int) $entryid, $data, true ) ) {
				continue;
			} else {
				foreach ( $entry as $addentry ) {
					if ( (int) $entryid === (int) $addentry['id'] ) {
						if ( $is_condition_enabled && ! empty( $logic ) ) {
							$check = GFCommon::evaluate_conditional_logic( $logic, $wpssg_form_data, $addentry );
							if ( ! $check ) {
								continue;
							}
						}
						if ( $is_creditcard_enabled ) {
							if ( isset( $settings['paymentstatus'] ) && $settings['paymentstatus'] ) {
								if ( ! in_array( $addentry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
									continue;
								}
							}
						}
						$newenrty[] = $addentry;
					}
				}
			}
		}
		$total_entries = count( $newenrty );
		if ( $total_entries > $entry_limit ) {
			$total_entries = $entry_limit;
		}
		$values_data = array();
		for ( $i = 0; $i < $total_entries; $i++ ) {
			$values_data[] = self::$instance->wpssg_prepared_values( $wpssg_form_data, $val, $newenrty[ $i ], $settings );
		}

		$response            = self::$instance_api->get_sheet_listing( $spreadsheetid );
		$existingsheetsnames = self::$instance_api->get_sheet_list( $response );

		$sheetid = $existingsheetsnames[ $range ];

		$startindex = count( $data ) + 1;
		$endindex   = $startindex + $total_entries;

		$param               = array();
		$param['sheetid']    = $sheetid;
		$param['startindex'] = $startindex;
		$param['endindex']   = $endindex;

		$batchupdaterequest = self::$instance_api->insertdimensionobject( $param );

		$requestobject                  = array();
		$requestobject['spreadsheetid'] = $spreadsheetid;
		$requestobject['requestbody']   = $batchupdaterequest;
		$response                       = self::$instance_api->formatsheet( $requestobject );
		if ( is_array( $values_data ) && ! empty( $values_data ) ) {
			try {
				/*Append entry*/
				$requestbody = self::$instance_api->valuerangeobject( $values_data );
				$params      = self::$instance->get_row_format( $settings );
				$range       = $range . '!A' . $startindex;
				$param       = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
				$response    = self::$instance_api->updateentry( $param );

			} catch ( Exception $e ) {
				$allowed_html = wp_kses_allowed_html( 'post' );
				echo wp_kses( 'Message : ' . $e->getMessage(), $allowed_html );
			}
		}
		echo esc_html( 'successful' );
		die;
	}
	/**
	 * Clear the Spreadsheet.
	 */
	public function wpssg_clear_spreadsheet() {
		if ( ! isset( $_POST['_wpnonce'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['_wpnonce'] ) ), 'save_general_settings' ) ) {
			echo esc_html__( 'Sorry, your nonce did not verify.', 'wpssg' );
			wp_die();
		}
		if ( isset( $_POST['form_id'] ) ) {
			$form_id = (int) sanitize_text_field( wp_unslash( $_POST['form_id'] ) );
		}
		if ( isset( $_POST['fid'] ) ) {
			$wpssg_fid = (int) sanitize_text_field( wp_unslash( $_POST['fid'] ) );
		}

		$syncformfeed    = RGFormsModel::get_form_meta( $form_id );
		$formfeed        = self::$instance->get_feeds( $form_id );
		$form_field_type = self::$instance->wpssg_form_field_type( $syncformfeed );

		foreach ( $formfeed as $feedsetting ) {
			if ( (int) $feedsetting['id'] === $wpssg_fid ) {
				$settings = $feedsetting['meta'];
			}
		}

		$spreadsheetid = $settings['spreedsheets'];
		$sheetname     = $settings['sheetselect'];
		$total         = self::$instance_api->get_row_list( $spreadsheetid, $sheetname );
		$total_headers = count( $total['values'][0] );
		$last_column   = self::$instance->wpssg_get_column_index( $total_headers );
		try {
			$range                  = $sheetname . '!A2:' . $last_column . '10000';
			$requestbody            = self::$instance_api->clearobject();
			$param                  = array();
			$param['spreadsheetid'] = $spreadsheetid;
			$param['sheetname']     = $range;
			$param['requestbody']   = $requestbody;
			$response               = self::$instance_api->clear( $param );
		} catch ( Exception $e ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( 'Message : ' . $e->getMessage(), $allowed_html );
		}

		echo esc_html( 'successful' );
		wp_die();
	}

	/**
	 * Return Character based on gven number
	 *
	 * @param int $number Number.
	 */
	public static function wpssg_get_column_index( $number ) {

		if ( $number <= 0 ) {
			return null;
		}

		$temp;
		$letter = '';
		while ( $number > 0 ) {
			$temp   = ( $number - 1 ) % 26;
			$letter = chr( $temp + 65 ) . $letter;
			$number = ( $number - $temp - 1 ) / 26;
		}
		return $letter;
	}

	/**
	 * Prepare Google Spreedsheet sheet name for feed setting.
	 *
	 * @access public
	 * @return array $choices
	 */
	public function wpssg_sheet_for_feed_setting() {

		/* Build choices array. */
		$choices = array(
			array(
				'label' => esc_html__( 'Select a Sheet List', 'wpssg' ),
				'value' => '',
			),
		);

		/* Get feed settings. */
		$settings = self::$instance->is_postback() ? self::$instance->get_posted_settings() : self::$instance->get_feed( self::$instance->get_current_feed_id() );

		$settings = isset( $settings['meta'] ) ? $settings['meta'] : $settings;
		if ( ! isset( $settings['spreedsheets'] ) || '' === (string) $settings['spreedsheets'] || 'new' === (string) $settings['spreedsheets'] ) {
			return $choices;
		}

		$response = self::$instance_api->get_sheet_listing( $settings['spreedsheets'] );

		foreach ( $response->getSheets() as $s ) {
			$sheets[] = $s['properties']['title'];
		}

		/* Add the Google Sheets to the choices array. */
		if ( ! empty( $sheets ) ) {
			$i = 0;
			foreach ( $sheets as $list ) {

				$choices[] = array(
					'label' => $list,
					'value' => $list,
				);
				$i++;
			}
		}
		return $choices;
	}

	/**
	 * Prepare Google Spreedsheet sheet headers for mapping fields to form fields.
	 *
	 * @access public
	 * @return array $choices
	 */
	public function wpssg_check_spreadsheet_header_value() {

		$settings = self::$instance->is_postback() ? self::$instance->get_posted_settings() : self::$instance->get_feed( self::$instance->get_current_feed_id() );
		$settings = isset( $settings['meta'] ) ? $settings['meta'] : $settings;

		if ( ! empty( $settings ) && isset( $settings['sheetselect'] ) ) {
			$sheet_data = self::$instance->wpssg_check_sheet_exist( $settings['spreedsheets'], $settings['sheetselect'] );
			if ( '' !== $sheet_data['spreadsheetid'] && '' !== $sheet_data['sheetname'] ) {
				$spreadsheetid = $settings['spreedsheets'];
				$sheetname     = $settings['sheetselect'] . '!A1:ZZ1';
				$response      = self::$instance_api->get_row_list( $spreadsheetid, $sheetname );
				$values        = $response->getValues();
				if ( empty( $response['values'] ) ) {
					return true;
				} else {
					return false;
				}
			}
		}
	}
	/**
	 * Check Google Spreedsheet sheet headers for mapping fields to form fields.
	 *
	 * @access public
	 * @return array $choices
	 */
	public function wpssg_check_spreadsheet_header_values() {

		$settings = self::$instance->is_postback() ? self::$instance->get_posted_settings() : self::$instance->get_feed( self::$instance->get_current_feed_id() );
		$settings = isset( $settings['meta'] ) ? $settings['meta'] : $settings;

		if ( ! empty( $settings ) && isset( $settings['sheetselect'] ) && isset( $settings['spreedsheets'] ) && ! empty( $settings['spreedsheets'] ) ) {
			$sheet_data = self::$instance->wpssg_check_sheet_exist( $settings['spreedsheets'], $settings['sheetselect'] );
			if ( '' !== $sheet_data['spreadsheetid'] && '' !== $sheet_data['sheetname'] ) {
				$spreadsheetid = $settings['spreedsheets'];
				$sheetname     = $settings['sheetselect'] . '!A1:ZZ1';
				$response      = self::$instance_api->get_row_list( $spreadsheetid, $sheetname );
				$values        = $response->getValues();

				if ( empty( $response['values'] ) ) {
					return false;
				} else {
					return true;
				}
			}
		}
	}
	/**
	 * Prepare Google Spreedsheet sheet headers for mapping fields to form fields.
	 *
	 * @access public
	 * @return array $choices
	 */
	public function wpssg_mapfield_for_feed_setting() {

		$settings = self::$instance->is_postback() ? self::$instance->get_posted_settings() : self::$instance->get_feed( self::$instance->get_current_feed_id() );
		$settings = isset( $settings['meta'] ) ? $settings['meta'] : $settings;
		//phpcs:ignore
		if ( isset( $_GET['id'] ) ) {
			//phpcs:ignore
			$form_id = (int) sanitize_text_field( wp_unslash( $_GET['id'] ) );
			$form    = RGFormsModel::get_form_meta( $form_id );
		}
		$fields_array = self::$instance->wpssg_generate_headers( $form, 'id' );
		if ( ! empty( $settings ) && isset( $settings['sheetselect'] ) ) {
			$sheet_data = self::$instance->wpssg_check_sheet_exist( $settings['spreedsheets'], $settings['sheetselect'] );
			if ( '' !== $sheet_data['spreadsheetid'] && '' !== $sheet_data['sheetname'] ) {
				$spreadsheetid = $settings['spreedsheets'];
				$sheetname     = $settings['sheetselect'] . '!A1:ZZ1';
				$response      = self::$instance_api->get_row_list( $spreadsheetid, $sheetname );
				$values        = $response->getValues();
				$choices       = array();
				if ( empty( $response['values'] ) ) {
					return;
				} else {
					$i = 0;
					foreach ( $values as $row ) {
						if ( 0 === (int) $i ) {
							foreach ( $row as $list ) {
								if ( 'Submission Date' === (string) $list || 'Payment Status' === (string) $list || 'Form ID' === (string) $list || 'Form Title' === (string) $list ) {
									continue;
								}

								$choices[] = array(
									'name'          => str_replace( array( ' ', '.' ), '_', strtolower( trim( $list ) ) ),
									'label'         => $list,
									'default_value' => array_search( $list, $fields_array, true ),
								);
							}
							return $choices;
						}
						$i++;
					}
				}
			} else {
				return;
			}
		} else {
			return;
		}
	}

	/**
	 * Prepare Google Spreedsheet sheet headers.
	 *
	 * @param string $spreadsheetid Google Spreadsheet id.
	 * @param string $sheetname Google Sheet Name.
	 * @access public
	 * @return array $row
	 */
	public function wpssg_get_mapfields( $spreadsheetid, $sheetname ) {
		$response = self::$instance_api->get_row_list( $spreadsheetid, $sheetname );
		$values   = $response->getValues();
		if ( empty( $values ) ) {
			print "No data found.\n";
		} else {
			$i = 0;
			foreach ( $values as $row ) {
				// Print columns A and E, which correspond to indices 0 and 4.
				if ( 0 === (int) $i ) {
					return $row;
				}
				$i++;
			}
		}
	}

	/**
	 * Update single entry from frontend.
	 *
	 * @param bool  $update_result return result.
	 * @param array $entry Entry Array.
	 * @param int   $form_id Form ID.
	 */
	public function wpssg_frontend_entry_updated( $update_result, $entry, $form_id ) {

		$form = GFAPI::get_form( $form_id );
		self::$instance->wpssg_entry_updated( $form, $entry['id'] );
		return $update_result;
	}

	/**
	 * Updates single Entry objects.
	 *
	 * @since  1.8
	 * @access public
	 *
	 * @uses GFAPI::update_entry()
	 * @param array $form Form Array.
	 * @param array $entry_id The Entry objects.
	 */
	public function wpssg_entry_updated( $form, $entry_id ) {
		$entry = GFAPI::get_entry( $entry_id );

		if ( ! self::$instance_api->getClient() ) {
			return;
		}
		// Get the plugin settings.
		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );
		// Access a specific setting e.g. an authToken.
		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) ) {

			$form_field_type       = array();
			$formfeed              = self::$instance->get_feeds( $form['id'] );
			$wpssg_form_data       = GFAPI::get_form( $form['id'] );
			$form_field_type       = self::$instance->wpssg_form_field_type( $form );
			$countformfeed         = ! is_wp_error( $formfeed ) ? count( $formfeed ) : 0;
			$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );
			for ( $i = 0;$i < $countformfeed; $i++ ) {

				if ( 1 === (int) $formfeed[ $i ]['is_active'] && isset( $formfeed[ $i ]['meta']['editcheckbox'] ) && $formfeed[ $i ]['meta']['editcheckbox'] ) {
					$feed                 = $formfeed[ $i ]['meta'];
					$flag                 = 1;
					$is_condition_enabled = rgar( $feed, 'feed_condition_conditional_logic' ) === '1';
					$logic                = rgars( $feed, 'feed_condition_conditional_logic_object/conditionalLogic' );

					$entry = GFAPI::get_entry( $entry_id );

					if ( $is_condition_enabled && ! empty( $logic ) ) {
						$check = GFCommon::evaluate_conditional_logic( $logic, $form, $entry );
						if ( ! $check ) {
							continue;
						}
					}
					if ( $is_creditcard_enabled ) {
						if ( isset( $feed['paymentstatus'] ) && $feed['paymentstatus'] ) {
							if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
								continue;
							}
						}
					}

					$spreadsheetid = $feed['spreedsheets'];
					$range         = $feed['sheetselect'];

					$rangetofind = $range . '!A:Z';
					$sheet_data  = self::$instance->wpssg_check_sheet_exist( $spreadsheetid, $range );
					if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
						continue;
					}
					$allentry = self::$instance_api->get_row_list( $spreadsheetid, $rangetofind );
					$data     = $allentry->getValues();

					$data = array_map(
						function( $wpssg_element ) {
							if ( isset( $wpssg_element['0'] ) ) {
								return $wpssg_element['0'];
							} else {
								return '';
							}
						},
						$data
					);
					$data = self::$instance->wpssg_convert_int( $data );
					$num  = array_search( (int) $entry_id, $data, true );

					if ( $num > 0 ) {

						$num++;
						$rangetoupdate    = $range . '!A' . $num;
						$values_data      = array();
						$formmapfieldsval = self::$instance->wpssg_get_mapfields( $spreadsheetid, $range );
						$val              = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
						$values_data      = self::$instance->wpssg_prepared_values( $form, $val, $entry, $feed );
						$values           = array( $values_data );

						$requestbody = self::$instance_api->valuerangeobject( $values );
						$params      = self::$instance->get_row_format( $feed );
						$param       = self::$instance_api->setparamater( $spreadsheetid, $rangetoupdate, $requestbody, $params );
						$response    = self::$instance_api->updateentry( $param );
					}
				}
			}
		} else {
			return;
		}
	}

	/** Conver string to readable form.
	 *
	 * @param array $chars Character.
	 * @access public
	 * @return array $row
	 */
	public function wpssg_convert_to_readable( $chars ) {

		$encoding = ini_get( 'mbstring.internal_encoding' );

		$str = preg_replace_callback(
			'/\\\\u([0-9a-fA-F]{4})/u',
			function( $match ) use ( $encoding ) {
				return mb_convert_encoding( pack( 'H*', $match[1] ), $encoding, 'UTF-16BE' );
			},
			$chars
		);

		return $str;
	}

	/**
	 * Admin Payment Notification.
	 *
	 * @param array  $feed Feed Array.
	 * @param array  $entry Entry Array.
	 * @param string $status Status.
	 * @param string $transaction_id Transaction ID.
	 * @param string $subscriber_id Subscriber ID.
	 * @param int    $amount Amount.
	 * @param string $pending_reason Pending Payment Reason.
	 * @param string $reason reason.
	 * @access public
	 */
	public function wpssg_admin_payment_notification( $feed, $entry, $status, $transaction_id, $subscriber_id, $amount, $pending_reason, $reason ) {
		$form = GFAPI::get_form( $entry['form_id'] );
		self::$instance->wpssg_update_payment_status( $entry['id'] );

	}
	/**
	 * Insert data into Google Spreadsheet after form submission.
	 *
	 * @param array $entry Entry Array.
	 * @param array $form Form Array.
	 * @access public
	 * @return array $row
	 */
	public function wpssg_after_submission( $entry, $form ) {

		if ( ! self::$instance_api->getClient() ) {
			return;
		}
		// phpcs:ignore
		if ( isset( $_POST['gpnf_mode'] ) && 'edit' === sanitize_text_field( wp_unslash( $_POST['gpnf_mode'] ) ) ) {
			self::$instance->wpssg_entry_updated( $form, $entry['id'] );
			return;
		}

		$form_field_type = array();

		$formfeed              = self::$instance->get_feeds( $form['id'] );
		$wpssg_form_data       = GFAPI::get_form( $form['id'] );
		$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );
		$form_field_type       = self::$instance->wpssg_form_field_type( $form );
		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );
		$formfeedcountdata     = ! is_wp_error( $formfeed ) ? count( $formfeed ) : 0;

		// Access a specific setting e.g. an authToken.
		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) ) {
			try {
				for ( $i = 0; $i < $formfeedcountdata; $i++ ) {
					$values_data = array();

					if ( '1' === (string) $formfeed[ $i ]['is_active'] ) {

						if ( $is_creditcard_enabled ) {
							if ( isset( $formfeed[ $i ]['meta']['paymentstatus'] ) && $formfeed[ $i ]['meta']['paymentstatus'] ) {
								if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
									continue;
								}
							}
						}
						$feed = $formfeed[ $i ]['meta'];
						$flag = 1;

						$is_condition_enabled = rgar( $feed, 'feed_condition_conditional_logic' ) === '1';

						$logic = rgars( $feed, 'feed_condition_conditional_logic_object/conditionalLogic' );

						if ( ! empty( $is_condition_enabled ) && ! empty( $logic ) ) {
							$check = GFCommon::evaluate_conditional_logic( $logic, $form, $entry );

							if ( ! $check ) {
								continue;
							}
						}

						if ( isset( $feed['spreedsheets'] ) && isset( $feed['sheetselect'] ) ) {
							$sheet_data = self::$instance->wpssg_check_sheet_exist( $feed['spreedsheets'], $feed['sheetselect'] );
							if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
								continue;
							}

							$formmapfieldsval[ $i ] = self::$instance->wpssg_get_mapfields( $feed['spreedsheets'], $feed['sheetselect'] );

							$spreadsheetid         = $feed['spreedsheets'];
							$range                 = $feed['sheetselect'];
							$countformmapfieldsval = count( $formmapfieldsval[ $i ] );

							$val         = self::$instance->wpssg_prepared_headers( $formmapfieldsval[ $i ], $feed );
							$values_data = self::$instance->wpssg_prepared_values( $form, $val, $entry, $feed );

							$values = array( $values_data );

							$allentry    = self::$instance_api->get_row_list( $spreadsheetid, $range . '!A:Z' );
							$data        = $allentry->getValues();
							$requestbody = self::$instance_api->valuerangeobject( $values );
							$params      = self::$instance->get_row_format( $feed );
							$range       = $range . '!A' . ( count( $data ) + 1 );
							$param       = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
							$response    = self::$instance_api->updateentry( $param );
						}
					}
				}
			} catch ( Exception $e ) {
				$allowed_html = wp_kses_allowed_html( 'post' );
				echo wp_kses( 'Message : ' . $e->getMessage(), $allowed_html );
			}
		}

	}

	/**
	 * Check is credit card field enable or not.
	 *
	 * @param array $form Form Array.
	 * @access public
	 * @return bool
	 */
	public function wpssg_get_credit_card_field( $form ) {
		$fields   = GFAPI::get_fields_by_type( $form, array( 'creditcard' ) );
		$response = empty( $fields ) ? false : $fields[0];
		if ( ! $response ) {
			$fields   = GFAPI::get_fields_by_type( $form, array( 'stripe_creditcard' ) );
			$response = empty( $fields ) ? false : $fields[0];
		}
		if ( ! $response ) {
			$is_mollie_active = self::$instance->wpssg_is_mollie_active();
			$redsysforgravity = self::$instance->wpssg_is_redsysforgravity_active();
			if ( $is_mollie_active || $redsysforgravity ) {
				$response = true;
			} else {
				$response = false;
			}
		}

		return empty( $response ) ? false : $response;
	}

	/**
	 * Prepare Google Spreadsheet list
	 *
	 * @access public
	 * @return array $sheetarray
	 */
	public function wpssg_list_googlespreedsheet() {
		$settings = self::$instance->is_postback() ? self::$instance->get_posted_settings() : self::$instance->get_feed( self::$instance->get_current_feed_id() );
		$settings = isset( $settings['meta'] ) ? $settings['meta'] : $settings;

		/* Build choices array. */
		$sheetarray = array(
			array(
				'label' => esc_html__( 'Select Google Spreeadsheet List', 'wpssg' ),
				'value' => '',
			),
		);
		$sheetarray = self::$instance_api->get_spreadsheet_listing( $sheetarray );
		return $sheetarray;
	}

	/**
	 * Save Feed Settings
	 *
	 * @param int   $feed_id Feed ID.
	 * @param int   $form_id Form ID.
	 * @param array $settings Settings Array.
	 * @access public
	 * @return array $sheetarray
	 */
	public function save_feed_settings( $feed_id, $form_id, $settings ) {
		if ( ! isset( $_POST['wpssg_general_settings'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['wpssg_general_settings'] ) ), 'save_general_settings' ) ) {
			return false;
		}
		if ( $feed_id ) {
			$result = self::$instance->update_feed_meta( $feed_id, $settings );
			$result = $feed_id;
		} else {
			$result = self::$instance->insert_feed( $form_id, true, $settings );
		}

		$spreadsheetid = isset( $_POST['_gaddon_setting_spreedsheets'] ) ? sanitize_text_field( wp_unslash( $_POST['_gaddon_setting_spreedsheets'] ) ) : '';
		if ( empty( $spreadsheetid ) ) {
			$spreadsheetid = isset( $_POST['_gform_setting_spreedsheets'] ) ? sanitize_text_field( wp_unslash( $_POST['_gform_setting_spreedsheets'] ) ) : '';
		}
		$sheetname = isset( $_POST['_gaddon_setting_sheetselect'] ) ? sanitize_text_field( wp_unslash( $_POST['_gaddon_setting_sheetselect'] ) ) : '';
		if ( empty( $sheetname ) ) {
			$sheetname = isset( $_POST['_gform_setting_sheetselect'] ) ? sanitize_text_field( wp_unslash( $_POST['_gform_setting_sheetselect'] ) ) : '';
		}

		/*
		* Check for create new spreadsheet request
		*
		*/
		if ( isset( $_POST['createnewspreadsheet'] ) ) {

			if ( isset( $_POST['newspreadsheetlabel'] ) && ! empty( $_POST['newspreadsheetlabel'] ) ) {
				$spreadsheetname = sanitize_text_field( wp_unslash( $_POST['newspreadsheetlabel'] ) );
			}

			if ( isset( $_POST['sheetlabel'] ) && ! empty( $_POST['sheetlabel'] ) ) {
				$sheetname = sanitize_text_field( wp_unslash( $_POST['sheetlabel'] ) );
			}
			// TODO: Assign values to desired properties of `requestbody`.
			$requestbody = self::$instance_api->newspreadsheetobject( $spreadsheetname, $sheetname );
			$response    = self::$instance_api->createspreadsheet( $requestbody );
			// phpcs:ignore
			$spreadsheetid = $response->spreadsheetId;
			$range         = trim( $sheetname ) . '!A1';
			$form          = RGFormsModel::get_form_meta( $form_id );

			$fields_array = self::$instance->wpssg_generate_headers( $form );

			$values = array( array_values( array_filter( $fields_array ) ) );

			$requestbody = self::$instance_api->valuerangeobject( $values );

			$params   = self::$instance->get_row_format();
			$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
			$response = self::$instance_api->appendentry( $param );

			$settings['spreedsheets'] = $spreadsheetid;
			$settings['sheetselect']  = $sheetname;

			$fid     = $result;
			$result1 = self::$instance->update_feed_meta( $result, $settings );
			// phpcs:ignore
			print( '<script>window.location.href="admin.php?page=gf_edit_forms&view=settings&subview=ggspreadsheetaddon&id=' . $form_id . '&fid=' . $fid . '"</script>' );
		}

		if ( ( isset( $_POST['_gaddon_setting_freezecheckbox'] ) && sanitize_text_field( wp_unslash( $_POST['_gaddon_setting_freezecheckbox'] ) ) !== '' ) || ( isset( $_POST['_gform_setting_freezecheckbox'] ) && sanitize_text_field( wp_unslash( $_POST['_gform_setting_freezecheckbox'] ) ) !== '' ) && 'new' !== $spreadsheetid ) {

			$wpssg_freeze = isset( $_POST['_gaddon_setting_freezecheckbox'] ) ? sanitize_text_field( wp_unslash( $_POST['_gaddon_setting_freezecheckbox'] ) ) : '';
			if ( empty( $wpssg_freeze ) || 0 === $wpssg_freeze ) {
				$wpssg_freeze = isset( $_POST['_gform_setting_freezecheckbox'] ) ? sanitize_text_field( wp_unslash( $_POST['_gform_setting_freezecheckbox'] ) ) : '';
			}

			if ( 'on' === (string) $wpssg_freeze || 1 === (int) $wpssg_freeze ) {
				$wpssg_freeze = 1;
			} else {
				$wpssg_freeze = 0;
			}
			$response = self::$instance_api->get_sheet_listing( $spreadsheetid );
			foreach ( $response->getSheets() as $key => $value ) {
				if ( (string) $value['properties']['title'] === $sheetname ) {
					$requestbody                    = self::$instance_api->freezeobject( $value['properties']['sheetId'], $wpssg_freeze );
					$requestobject                  = array();
					$requestobject['spreadsheetid'] = $spreadsheetid;
					$requestobject['requestbody']   = $requestbody;
					self::$instance_api->formatsheet( $requestobject );
				}
			}
		}
		if ( isset( $_POST['submissiondate'] ) ) {
			$range    = $sheetname . '!A1:ZZ1';
			$response = self::$instance_api->get_row_list( $spreadsheetid, $range );
			$data     = $response->getValues();

			if ( ! in_array( 'Submission Date', $data[0], true ) ) {
				$fields_array   = $data[0];
				$fields_array[] = 'Submission Date';
				$range          = $sheetname . '!A1';
				$values         = array( $fields_array );

				$requestbody = self::$instance_api->valuerangeobject( $values );
				$params      = self::$instance->get_row_format();
				$param       = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
				$response    = self::$instance_api->updateentry( $param );
			}
		}
		if ( isset( $_POST['paymentstatusheader'] ) ) {
			$range    = $sheetname . '!A1:ZZ1';
			$response = self::$instance_api->get_row_list( $spreadsheetid, $range );
			$data     = $response->getValues();

			if ( ! in_array( 'Payment Status', $data[0], true ) ) {
				$fields_array   = $data[0];
				$fields_array[] = 'Payment Status';
				$range          = $sheetname . '!A1';
				$values         = array( $fields_array );
				$requestbody    = self::$instance_api->valuerangeobject( $values );
				$params         = self::$instance->get_row_format();
				$param          = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
				$response       = self::$instance_api->updateentry( $param );

			}
		}
		if ( isset( $_POST['wpssg_form_id'] ) ) {
			$range    = $sheetname . '!A1:ZZ1';
			$response = self::$instance_api->get_row_list( $spreadsheetid, $range );
			$data     = $response->getValues();
			if ( ! in_array( 'Form ID', $data[0], true ) ) {
				$fields_array   = $data[0];
				$fields_array[] = 'Form ID';
				$range          = $sheetname . '!A1';
				$values         = array( $fields_array );
				$requestbody    = self::$instance_api->valuerangeobject( $values );
				$params         = self::$instance->get_row_format();
				$param          = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
				$response       = self::$instance_api->updateentry( $param );
			}
		}
		if ( isset( $_POST['wpssg_form_title'] ) ) {
			$range    = $sheetname . '!A1:ZZ1';
			$response = self::$instance_api->get_row_list( $spreadsheetid, $range );
			$data     = $response->getValues();

			if ( ! in_array( 'Form Title', $data[0], true ) ) {
				$fields_array   = $data[0];
				$fields_array[] = 'Form Title';
				$range          = $sheetname . '!A1';
				$values         = array( $fields_array );

				$requestbody = self::$instance_api->valuerangeobject( $values );
				$params      = self::$instance->get_row_format();

				$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
				$response = self::$instance_api->updateentry( $param );
			}
		}
		$_POST['form_id'] = $form_id;
		/* Schedule Auto Sync */
		if ( isset( $_POST['scheduling_enable'] ) ) {
			$schedule_data = self::$instance->wpssg_update_schedule_data( $_POST, $result, $settings );
			if ( ! is_wp_error( $schedule_data ) ) {
				$settings = $schedule_data;
			}
		}
		/* Schedule Auto Import */
		if ( isset( $_POST['import_scheduling_enable'] ) ) {
			$import_schedule_data = self::$instance->wpssg_update_import_schedule_data( $_POST, $result, $settings );
			if ( ! is_wp_error( $import_schedule_data ) ) {
				$settings = $import_schedule_data;
			}
		}
		if ( isset( $_POST['headerset'] ) ) {
			$range = $sheetname . '!A1';

			$form         = RGFormsModel::get_form_meta( $form_id );
			$fields_array = self::$instance->wpssg_generate_headers( $form );
			$values       = array( array_values( array_filter( $fields_array ) ) );

			$requestbody = self::$instance_api->valuerangeobject( $values );

			$params   = self::$instance->get_row_format();
			$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
			$response = self::$instance_api->updateentry( $param );

			$fid = $result;
			// phpcs:ignore
			print( '<script>window.location.href="admin.php?page=gf_edit_forms&view=settings&subview=ggspreadsheetaddon&id=' . $form_id . '&fid=' . $fid . '"</script>' );
		}
		return $result;
	}

	/**
	 * Enable feed duplication.
	 *
	 * @param int $id ID.
	 * @access public
	 * @return bool
	 */
	public function can_duplicate_feed( $id ) {
		return true;
	}

	/**
	 * Heading row for field map table.
	 *
	 * @since  2.2
	 * @access public
	 *
	 * @uses GFAddOn::field_map_title()
	 *
	 * @return string
	 */
	public function field_map_table_header() {
		return '<thead>
			<tr>
				<th>' . esc_html__( 'Sheet Headers', 'wpssg' ) . '</th>
				<th>' . esc_html__( 'Form Field', 'wpssg' ) . '</th>
			</tr>
		</thead>';
	}

	/**
	 * Generate headers.
	 *
	 * @param array  $form Form Array.
	 * @param string $format Return with id.
	 */
	public function wpssg_generate_headers( $form, $format = '' ) {

		$fields_array       = array();
		$fields_array['id'] = 'Entry Id';
		if ( is_array( $form['fields'] ) ) {
			foreach ( $form['fields'] as $field ) {
				if ( 1 === (int) $field['displayOnly'] ) {
					continue;
				}
				if ( 'checkbox' === (string) $field['type'] ) {
					$fields_array[ $field->id ] = GFCommon::get_label( $field, $field->id );
				} elseif ( isset( $field['inputs'] ) && is_array( $field['inputs'] ) && ! empty( $field['inputs'] ) ) {

					foreach ( $field['inputs'] as $input ) {
						if ( ! isset( $input['isHidden'] ) ) {
							$fields_array[ $input['id'] ] = GFCommon::get_label( $field, $input['id'] );
						} elseif ( isset( $input['isHidden'] ) && 1 !== $input['isHidden'] ) {
							$fields_array[ $input['id'] ] = GFCommon::get_label( $field, $input['id'] );
						}
					}
				}

				if ( 'list' === (string) $field['type'] && is_array( $field['choices'] ) && ! empty( $field['choices'] ) ) {
					foreach ( $field['choices'] as $inx => $choice ) {
						if ( ! empty( $choice['text'] ) ) {
							$fields_array[ $field->id . '.' . $inx ] = $choice['text'];
						}
					}
				} elseif ( isset( $field ) && is_object( $field ) && ! isset( $field['inputs'] ) ) {
					$fields_array[ $field->id ] = GFCommon::get_label( $field, $field->id );
				}

				if ( 'fileupload' === (string) $field['type'] || 'dropbox' === (string) $field['type'] ) {
					$fileupload_fields = apply_filters( 'wpssg_generate_fileupload_headers', array(), $field );
					if ( is_array( $fileupload_fields ) && ! empty( $fileupload_fields ) ) {
						$fields_array = $fields_array + $fileupload_fields;
					}
				}
			}
			$is_mollie_active           = self::$instance->wpssg_is_mollie_active();
			$is_redsysforgravity_active = self::$instance->wpssg_is_redsysforgravity_active();
			if ( $is_mollie_active || $is_redsysforgravity_active ) {
				$fields_array['payment_date']   = 'Payment Date';
				$fields_array['payment_amount'] = 'Payment Amount';

				$fields_array['payment_method'] = 'Payment Method';
				$fields_array['transaction_id'] = 'Transaction ID';
			}
			$is_gpdfapi_active = self::$instance->wpssg_is_gpdfapi_active();
			if ( $is_gpdfapi_active ) {
				$fields_array['pdf'] = 'PDF';
			}
		}
		if ( 'id' === (string) $format ) {
			return $fields_array;
		} else {
			return array_values( $fields_array );
		}
	}

	/**
	 * Check Field is display only or not.
	 *
	 * @param array $form Form Array.
	 */
	public function wpssg_form_field_type( $form ) {
		$fields_array = array();
		if ( is_array( $form['fields'] ) ) {
			foreach ( $form['fields'] as $field ) {
				if ( 1 === (int) $field['displayOnly'] ) {
					continue;
				}
				$fields_array[ $field->id ] = $field['type'];
			}
		}
		return $fields_array;
	}
	/**
	 * Reset Google API Settings
	 */
	public static function wpssg_reset_settings() {
		try {
			$wpssg_google_settings_value = self::$instance_api->wpssg_option( 'wpssg_google_settings' );
			$settings                    = array();
			foreach ( $wpssg_google_settings_value as $key => $value ) {
				$settings[ $key ] = '';
			}
			self::$instance_api->wpssg_update_option( 'wpssg_google_settings', $settings );
			self::$instance_api->wpssg_update_option( 'wpssg_google_accessToken', '' );
		} catch ( Exception $e ) {
			$allowed_html = wp_kses_allowed_html( 'post' );
			echo wp_kses( $e->getMessage(), $allowed_html );
		}
		echo 'successful';
		wp_die();
	}

	/**
	 * Reset Google API Settings
	 */
	public static function wpssg_getdate_preview() {
		if ( ! isset( $_POST['sync_nonce_token'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['sync_nonce_token'] ) ), 'sync_nonce' )
		) {
			wp_die();
		}
		if ( isset( $_POST['dateformat'] ) ) {
			try {
				$wpssg_format = sanitize_text_field( wp_unslash( $_POST['dateformat'] ) );
				echo esc_html( wp_date( $wpssg_format ) );
			} catch ( Exception $e ) {
				$allowed_html = wp_kses_allowed_html( 'post' );
				echo wp_kses( $e->getMessage(), $allowed_html );
			}
		}
		wp_die();
	}

	/**
	 * Add WPSynSheets Column.
	 *
	 * @param array $table_columns Entry Listing Table Columns.
	 */
	public function wpssg_set_custom_columns( $table_columns ) {
		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );

		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) ) {

			if ( self::$instance_api->checkcredenatials() ) {
				// phpcs:ignore
				if ( ! isset( $_GET['id'] ) || (isset( $_GET['filter'] ) && 'trash' === (string) $_GET['filter'] ) ) {
					return $table_columns;
				}
				// phpcs:ignore
				$formdata      = GFAPI::get_form( sanitize_text_field( wp_unslash( $_GET['id'] ) ) );
				// phpcs:ignore
				$formfeed      = GFAPI::get_feeds( '', sanitize_text_field( wp_unslash( $_GET['id'] ) ) );

				if ( ! is_wp_error( $formfeed ) && count( $formfeed ) > 0 ) {
					$countformfeed = count( $formfeed );
					for ( $i = 0;$i < $countformfeed; $i++ ) {
						if ( 1 === (int) $formfeed[ $i ]['is_active'] ) {
							$settings = $formfeed[ $i ]['meta'];
							if ( isset( $settings['spreedsheets'] ) && isset( $settings['sheetselect'] ) && '' !== $settings['spreedsheets'] && '' !== $settings['sheetselect'] ) {
								$wpssg_column = self::$instance->wpssg_convert_string( array_keys( $table_columns ) );
								$key_index    = array_search( 'column_selector', $wpssg_column, true );
								$arrayend     = array_splice( $table_columns, $key_index );
								$arraystart   = array_splice( $table_columns, 0, $key_index );

								$table_columns = array_merge( $arraystart, array( 'field_id-sync_button' => 'WPSyncSheets' ), $arrayend );

								return $table_columns;
							}
						}
					}
				}
			}
		}
		return $table_columns;
	}

	/**
	 * Add Click to Sync button column in entry listing page.
	 *
	 * @param string $value Button HTML.
	 * @param string $form_id Form ID.
	 * @param string $field_id Field ID.
	 * @param array  $entry Entry .
	 * @param string $query_string Query String.
	 */
	public function wpssg_change_column_data( $value, $form_id, $field_id, $entry, $query_string ) {
		if ( 'sync_button' !== (string) $field_id ) {
			return $value;
		}
		$html  = sprintf(
			'<a href="#" class="button gform_google_single_entry_sync_btn">%1$s</a>',
			esc_html__( 'Click to Sync', 'wpssg' )
		);
		$html .= sprintf(
			'<img src="%1$s" class="syncbtnloader">',
			esc_url( admin_url( 'images/spinner.gif' ), 'wpssg' )
		);

		return $html;
	}

	/**
	 * Sync single entry data.
	 */
	public function wpssg_sync_single_entry_data() {
		if ( ! isset( $_POST['sync_nonce_token'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['sync_nonce_token'] ) ), 'sync_nonce' )
		) {
			wp_die();
		}

		if ( isset( $_POST['entry_id'] ) ) {
			$entry_id = (int) sanitize_text_field( wp_unslash( $_POST['entry_id'] ) );
		}

		if ( isset( $_POST['form_id'] ) ) {
			$form_id = sanitize_text_field( wp_unslash( $_POST['form_id'] ) );
		}

		if ( ! self::$instance_api->getClient() ) {
			return;
		}
		// Get the plugin settings.
		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );
		// Access a specific setting e.g. an authToken.
		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) ) {

			$form_field_type       = array();
			$form                  = GFAPI::get_form( $form_id );
			$formfeed              = self::$instance->get_feeds( $form_id );
			$wpssg_form_data       = GFAPI::get_form( $form_id );
			$form_field_type       = self::$instance->wpssg_form_field_type( $form );
			$countformfeed         = ! is_wp_error( $formfeed ) ? count( $formfeed ) : 0;
			$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $form );

			for ( $i = 0;$i < $countformfeed; $i++ ) {

				if ( 1 === (int) $formfeed[ $i ]['is_active'] ) {
					$entry = GFAPI::get_entry( $entry_id );

					if ( $is_creditcard_enabled ) {
						if ( isset( $formfeed[ $i ]['meta']['paymentstatus'] ) && $formfeed[ $i ]['meta']['paymentstatus'] ) {
							if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
								continue;
							}
						}
					}
					$feed = $formfeed[ $i ]['meta'];

					$flag                 = 1;
					$is_condition_enabled = rgar( $feed, 'feed_condition_conditional_logic' ) === '1';
					$logic                = rgars( $feed, 'feed_condition_conditional_logic_object/conditionalLogic' );

					if ( $is_condition_enabled && ! empty( $logic ) ) {
						$check = GFCommon::evaluate_conditional_logic( $logic, $form, $entry );
						if ( ! $check ) {
							continue;
						}
					}
					if ( ! isset( $feed['spreedsheets'] ) || empty( $feed['spreedsheets'] ) || ! isset( $feed['sheetselect'] ) || empty( $feed['sheetselect'] ) ) {
						continue;
					}
					$spreadsheetid = $feed['spreedsheets'];
					$range         = $feed['sheetselect'];

					$sheet_data = self::$instance->wpssg_check_sheet_exist( $feed['spreedsheets'], $feed['sheetselect'] );
					if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
						continue;
					}
					$response                  = self::$instance_api->get_sheet_listing( $spreadsheetid );
					$wpssg_existingsheetsnames = self::$instance_api->get_sheet_list( $response );

					$sheetid = $wpssg_existingsheetsnames[ $range ];

					$allentry = self::$instance_api->get_row_list( $spreadsheetid, $range . '!A:Z' );
					$data     = $allentry->getValues();

					$data = array_map(
						function( $element ) {
							if ( isset( $element['0'] ) ) {
								return $element['0'];
							} else {
								return '';
							}
						},
						$data
					);

					$values_data      = array();
					$formmapfieldsval = self::$instance->wpssg_get_mapfields( $spreadsheetid, $range );

					$countformmapfieldsval = count( $formmapfieldsval );
					$val                   = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
					$values_data           = self::$instance->wpssg_prepared_values( $form, $val, $entry, $feed );
					$values                = array( $values_data );

					$requestbody = self::$instance_api->valuerangeobject( $values );

					$params = self::$instance->get_row_format( $feed );
					$data   = self::$instance->wpssg_convert_int( $data );

					$num = array_search( $entry_id, $data, true );

					if ( $num > 0 ) {
							$num++;
							$range    = $range . '!A' . $num;
							$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
							$response = self::$instance_api->updateentry( $param );
					} else {
						$wpssg_append = 0;
						foreach ( $data as $wpssg_key => $wpssg_value ) {
							if ( ! empty( $wpssg_value ) ) {
								if ( ( (int) $entry_id < (int) $wpssg_value ) ) {

									$wpssg_append = 1;

									$startindex = $wpssg_key;
									$endindex   = $wpssg_key + 1;

									$param               = array();
									$param['sheetid']    = $sheetid;
									$param['startindex'] = $startindex;
									$param['endindex']   = $endindex;

									$batchupdaterequest = self::$instance_api->insertdimensionobject( $param );

									$requestobject                  = array();
									$requestobject['spreadsheetid'] = $spreadsheetid;
									$requestobject['requestbody']   = $batchupdaterequest;
									$response                       = self::$instance_api->formatsheet( $requestobject );
									$startindex++;
									$range    = $range . '!A' . $startindex;
									$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
									$response = self::$instance_api->updateentry( $param );
									break;

								}
							}
						}
						if ( 0 === (int) $wpssg_append ) {
							$wpssg_isupdated = 0;

							$requestbody = self::$instance_api->valuerangeobject( $values );
							$params      = self::$instance->get_row_format( $feed );

							if ( count( $data ) > 1 ) {

								$range = $range . '!A' . ( count( $data ) + 1 );

								$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
								$response = self::$instance_api->updateentry( $param );

								$wpssg_isupdated = 1;
							}
							if ( 0 === (int) $wpssg_isupdated ) {
								$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
								$response = self::$instance_api->appendentry( $param );
							}
						}
					}
				}
			}
			echo esc_html( 'successful' );
			wp_die();
		} else {
			return;
		}
	}
	/**
	 * Add the meta box to the entry detail page.
	 *
	 * @param array $meta_boxes The properties for the meta boxes.
	 * @param array $entry The entry currently being viewed/edited.
	 * @param array $form The form object used to process the current entry.
	 *
	 * @return array
	 */
	public function wpssg_entry_page_syncbtn( $meta_boxes, $entry, $form ) {

		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );

		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) && self::$instance_api->getClient() ) {

				$formdata = GFAPI::get_form( $form['id'] );
				$formfeed = GFAPI::get_feeds( '', $form['id'] );

			if ( ! is_wp_error( $formfeed ) && count( $formfeed ) > 0 ) {
				$formfeedcount = count( $formfeed );
				for ( $i = 0;$i < $formfeedcount; $i++ ) {
					if ( 1 === (int) $formfeed[ $i ]['is_active'] ) {
						$settings = $formfeed[ $i ]['meta'];
						if ( isset( $settings['spreedsheets'] ) && isset( $settings['sheetselect'] ) && '' !== $settings['spreedsheets'] && '' !== $settings['sheetselect'] ) {
							$meta_boxes[ $this->_slug ] = array(
								'title'    => $this->get_short_title(),
								'callback' => array( $this, 'wpssg_add_details_meta_box' ),
								'context'  => 'side',
							);
						}
					}
				}
			}
		}
		return $meta_boxes;
	}

	/**
	 * The callback used to echo the content to the meta box.
	 *
	 * @param array $args An array containing the form and entry objects.
	 */
	public function wpssg_add_details_meta_box( $args ) {
		$form  = $args['form'];
		$entry = $args['entry'];

		$html         = '';
		$html        .= sprintf(
			'<a href="#" id="gform_metabox_sync_btn" class="button gform_google_single_entry_sync_btn">%1$s</a>',
			esc_html__( 'Click to Sync', 'wpssg' )
		);
		$html        .= sprintf(
			'<img src="%1$s" id="gform_metabox_syncbtnloader" class="syncbtnloader">',
			esc_url( admin_url( 'images/spinner.gif' ), 'wpssg' )
		);
		$allowed_html = wp_kses_allowed_html( 'post' );
		echo wp_kses( $html, $allowed_html );
	}

	/**
	 * Change the row format of spreadsheet.
	 *
	 * @param array $settings An array containing the sheet object to change the row format of spreadsheet.
	 */
	public function get_row_format( $settings = array() ) {
		if ( isset( $settings['row_format'] ) && 'userenter' === (string) $settings['row_format'] ) {
			$params = array( 'valueInputOption' => 'USER_ENTERED' );
		} else {
			$params = array( 'valueInputOption' => 'RAW' );
		}
		return $params;
	}

	/**
	 * Prepare Header List.
	 *
	 * @param array $formmapfieldsval An array containing the form fields.
	 * @param array $feed An array form.
	 */
	public function wpssg_prepared_headers( $formmapfieldsval, $feed ) {
		$countformmapfieldsval = count( $formmapfieldsval );
		$val                   = array();
		for ( $j = 0; $j < $countformmapfieldsval; $j++ ) {
			if ( 'Submission Date' === (string) $formmapfieldsval[ $j ] ) {
				$val[ $j ] = 'Submission Date';
				continue;
			}
			if ( 'Payment Status' === (string) $formmapfieldsval[ $j ] ) {
				$val[ $j ] = 'Payment Status';
				continue;
			}
			if ( 'Form ID' === (string) $formmapfieldsval[ $j ] ) {
				$val[ $j ] = 'Form ID';
				continue;
			}
			if ( 'Form Title' === (string) $formmapfieldsval[ $j ] ) {
				$val[ $j ] = 'Form Title';
				continue;
			}
			$fieldname = str_replace( array( ' ', '.' ), '_', strtolower( trim( $formmapfieldsval[ $j ] ) ) );

			$custom_header = apply_filters( 'wpssg_headers_to_bypass', array() );
			$custom_header = array_filter( $custom_header );
			$custom_header = array_map(
				function( $element ) {
					return str_replace( array( ' ', '.' ), '_', strtolower( trim( $element ) ) );
				},
				$custom_header
			);

			if ( ! empty( $custom_header ) ) {
				if ( isset( $feed[ 'mappedFields_' . $fieldname ] ) && empty( $feed[ 'mappedFields_' . $fieldname ] ) && in_array( $fieldname, $custom_header, true ) ) {
					$val[ $j ] = $fieldname;
					continue;
				}
			}

			if ( isset( $feed[ 'mappedFields_' . $fieldname ] ) ) {
				$val[ $j ] = $feed[ 'mappedFields_' . $fieldname ];
			}
		}
		return $val;
	}

	/**
	 * Prepare Value for insert in Google Spreadsheet.
	 *
	 * @param array $form An array form.
	 * @param array $val An array containing the form fields.
	 * @param array $entry Gravity Form Entry.
	 * @param array $feed Feed Array.
	 *
	 * @return array.
	 */
	public function wpssg_prepared_values( $form, $val, $entry, $feed ) {

		$last_key        = '';
		$countval        = count( $val );
		$wpssg_form_data = GFAPI::get_form( $form['id'] );
		$form_field_type = self::$instance->wpssg_form_field_type( $form );
		$values_data     = array();
		$location        = array();
		if ( isset( $entry['ip'] ) && ! empty( $entry['ip'] ) ) {
			// phpcs:ignore
			$location_value  = file_get_contents( 'http://ipinfo.io/' . $entry['ip'] . '/geo' );
			$location       = json_decode( $location_value, true );
		}

		$countries_field = new GF_Field_Address();
		$countries       = $countries_field->get_default_countries();
		$browser_data    = array();
		if ( isset( $entry['user_agent'] ) && ! empty( $entry['user_agent'] ) ) {
			$browser      = new foroco\BrowserDetection();
			$browser_data = $browser->getAll( $entry['user_agent'] );
		}
		$pdf_values        = array();
		$is_gpdfapi_active = self::$instance->wpssg_is_gpdfapi_active();
		if ( in_array( 'pdf', $val, true ) && $is_gpdfapi_active ) {
			$pdf_preview_link = get_site_url() . '/pdf/{pdf_preview_id}/' . $entry['id'];

			$pdfs = GPDFAPI::get_entry_pdfs( $entry['id'] );

			if ( ! is_wp_error( $pdfs ) && count( $pdfs ) > 0 ) {
				$pdf_links = array_column( $pdfs, 'id' );

				foreach ( $pdf_links as $pdf_link ) {
					$pdf_values[] = str_replace( '{pdf_preview_id}', $pdf_link, $pdf_preview_link );
				}
			}
		}
		$custom_header = apply_filters( 'wpssg_headers_to_bypass', array() );
		$custom_header = array_filter( $custom_header );
		$custom_header = array_map(
			function( $element ) {
				return str_replace( array( ' ', '.' ), '_', strtolower( trim( $element ) ) );
			},
			$custom_header
		);

		for ( $k = 0; $k < $countval; $k++ ) {
			if ( ! empty( $custom_header ) && in_array( $val[ $k ], $custom_header, true ) ) {
				$values_data[ $k ] = Google_Model::NULL_VALUE;
				continue;
			}

			if ( 'Payment Status' === (string) $val[ $k ] && isset( $feed['paymentstatusheader'] ) && $feed['paymentstatusheader'] ) {
				$values_data[ $k ] = isset( $entry['payment_status'] ) ? $entry['payment_status'] : '';
				continue;
			}
			if ( 'Form ID' === (string) $val[ $k ] && isset( $feed['wpssg_form_id'] ) && $feed['wpssg_form_id'] ) {
				$values_data[ $k ] = isset( $form['id'] ) ? $form['id'] : '';
				continue;
			}
			if ( 'Form Title' === (string) $val[ $k ] && isset( $feed['wpssg_form_title'] ) && $feed['wpssg_form_title'] ) {
				$values_data[ $k ] = isset( $wpssg_form_data['title'] ) ? $wpssg_form_data['title'] : '';
				continue;
			}
			if ( 'Submission Date' === (string) $val[ $k ] && isset( $feed['submissiondate'] ) && $feed['submissiondate'] ) {
				if ( isset( $feed['submissiondate'] ) && ! empty( $feed['submissiondate'] ) ) {
					if ( isset( $feed['submission_date_format'] ) && ! empty( $feed['submission_date_format'] ) ) {
						$values_data[ $k ] = $this->wpssg_get_date( $entry['date_created'], $feed['submission_date_format'] );
					} else {
						$values_data[ $k ] = $this->wpssg_get_date( $entry['date_created'], $feed['submissiondate_format'] );
					}
				} else {
					$values_data[ $k ] = '';
				}
				continue;
			}
			$num        = $val[ $k ];
			$is_explode = 0;
			if ( false !== strpos( $num, '.' ) ) {
				$fid        = explode( '.', $num );
				$num        = $fid[0];
				$is_explode = 1;
			}

			if ( ! empty( $num ) && array_key_exists( $num, $form_field_type ) ) {
				if ( 'checkbox' === (string) $form_field_type[ $num ] && 0 === (int) $is_explode ) {
					$field          = GFFormsModel::get_field( $form, $num );
					$checkbox_value = '';
					foreach ( $field['inputs'] as $choice ) {
						if ( ! empty( $entry[ $choice['id'] ] ) ) {
							$checkbox_value .= $entry[ $choice['id'] ] . ',';
						}
					}
					$values_data[ $k ] = rtrim( $checkbox_value, ',' );
					continue;
				} elseif ( 'list' === (string) $form_field_type[ $num ] ) {
					if ( '' === (string) $entry[ $num ] || null === $entry[ $num ] ) {
						$values_data[ $k ] = '';
					} else {
						$last_key         = $num;
						$wpssg_form_field = GFFormsModel::get_field( $wpssg_form_data, $num );
						$is_enablecolumns = isset( $wpssg_form_field['enableColumns'] ) ? $wpssg_form_field['enableColumns'] : 0;
						// phpcs:ignore
						$list_array = array();
						$list_array = maybe_unserialize( $entry[ $num ] );
						if ( isset( $list_array[0] ) && ! empty( $list_array[0] ) && 1 === (int) $is_enablecolumns ) {
							if ( 0 === (int) $is_explode ) {
								$list_row_array = array();
								foreach ( $list_array as $list_rows ) {
									$list_row_array[] = implode( ',', $list_rows );
								}
								$list_row          = implode( '|', $list_row_array );
								$values_data[ $k ] = $list_row;
							} else {
								$listkeys       = array_keys( $list_array[0] );
								$list_row_array = array();

								$list_columns = array();
								foreach ( $wpssg_form_field['choices'] as $choice ) {
									$list_columns[] = isset( $choice['value'] ) ? $choice['value'] : '';
								}

								foreach ( $list_columns as $ldata ) {
									$list_row_array[] = in_array( $ldata, $listkeys, true ) ? implode( ',', array_column( $list_array, $ldata ) ) : '';
								}
								if ( false !== strpos( $val[ $k ], '.' ) ) {
									$list_field_id     = explode( '.', $val[ $k ] );
									$values_data[ $k ] = isset( $list_row_array[ $list_field_id[1] ] ) ? $list_row_array[ $list_field_id[1] ] : '';
								}
							}
						} elseif ( ! empty( $list_array ) ) {
							$values_data[ $k ] = implode( ',', $list_array );
						} else {
							$values_data[ $k ] = '';
						}
					}
					continue;
				} elseif ( 'name' === (string) $form_field_type[ $num ] && 0 === (int) $is_explode ) {
					$values_data[ $k ] = $this->get_full_name( $entry, $num );
					continue;
				} elseif ( 'address' === (string) $form_field_type[ $num ] && 0 === (int) $is_explode ) {
					$address_arr = array();
					$address     = '';
					foreach ( $entry as $field_key => $field_val ) {
						if ( 1 === strpos( '"' . $field_key, $num . '.' ) ) {
							$address_arr[ $field_key ] = $field_val;
						}
					}
					$address_arr = array_filter( $address_arr );
					ksort( $address_arr );
					$address           = implode( ',', $address_arr );
					$values_data[ $k ] = $address;
					continue;
				} elseif ( 'fileupload' === (string) $form_field_type[ $num ] && 0 === (int) $is_explode ) {
					if ( isset( $entry[ $val[ $k ] ] ) && ( ! empty( $entry[ $val[ $k ] ] ) ) ) {
						$wpssg_url     = json_decode( $entry[ $val[ $k ] ] );
						$wpssg_url_val = '';
						if ( null === $wpssg_url ) { // check if it was invalid json string.
							$wpssg_url_val = $entry[ $val[ $k ] ];
						}

						if ( is_array( $wpssg_url ) ) {
							$wpssg_url_val = implode( ',', $wpssg_url );
						}
						$wpssg_url_val     = str_replace( array( '[', ']', '"' ), '', $wpssg_url_val );
						$values_data[ $k ] = $wpssg_url_val;
					} else {
						$values_data[ $k ] = '';
					}
					continue;
				} elseif ( ( 'fileupload' === (string) $form_field_type[ $num ] || 'dropbox' === (string) $form_field_type[ $num ] ) && 1 === (int) $is_explode ) {
					$values_data[ $k ] = apply_filters( 'wpssg_prepare_fileupload_values', '', $val[ $k ], $entry[ $num ] );
					continue;
				}
			}

			if ( 'form_title' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $wpssg_form_data['title'] ) ? $wpssg_form_data['title'] : '';
				continue;
			} elseif ( 'ip_address' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $entry['ip'] ) ? $entry['ip'] : '';
				continue;
			} elseif ( 'city' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $location['city'] ) ? $location['city'] : '';
				continue;
			} elseif ( 'state' === (string) $val[ $k ] ) {
				if ( isset( $location['region'] ) ) {
					$values_data[ $k ] = $location['region'];
				} elseif ( isset( $location['state'] ) ) {
					$values_data[ $k ] = $location['state'];
				} else {
					$values_data[ $k ] = '';
				}
				continue;
			} elseif ( 'country' === (string) $val[ $k ] || 'country_code' === (string) $val[ $k ] ) {
				$country = isset( $location['country'] ) ? $location['country'] : '';
				if ( 'country' === (string) $val[ $k ] && ! empty( $country ) ) {
					$country = isset( $countries[ $country ] ) ? $countries[ $country ] : '';
				}
				$values_data[ $k ] = $country;
				continue;
			} elseif ( 'timezone' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $location['timezone'] ) ? $location['timezone'] : '';
				continue;
			} elseif ( 'browser_name' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $browser_data['browser_title'] ) ? $browser_data['browser_title'] : '';
				continue;
			} elseif ( 'operating_system' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $browser_data['os_title'] ) ? $browser_data['os_title'] : '';
				continue;
			} elseif ( 'device_type' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $browser_data['device_type'] ) ? $browser_data['device_type'] : '';
				continue;
			} elseif ( 'payment_method' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $entry['payment_method'] ) ? $entry['payment_method'] : '';
				continue;
			} elseif ( 'transaction_id' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $entry['transaction_id'] ) ? $entry['transaction_id'] : '';
				continue;
			} elseif ( 'payment_amount' === (string) $val[ $k ] ) {
				$values_data[ $k ] = isset( $entry['payment_amount'] ) ? $entry['payment_amount'] : '';
				continue;
			} elseif ( 'pdf' === (string) $val[ $k ] ) {
				$values_data[ $k ] = implode( ',', $pdf_values );
				continue;
			} elseif ( ! isset( $entry[ $val[ $k ] ] ) || '' === (string) $entry[ $val[ $k ] ] || null === $entry[ $val[ $k ] ] ) {
				$values_data[ $k ] = '';
				continue;
			} else {
				if ( ( isset( $form_field_type[ $num ] ) && 'date' === (string) $form_field_type[ $num ] ) || 'date_created' === (string) $val[ $k ] || 'payment_date' === (string) $val[ $k ] ) {
					if ( isset( $feed['submission_date_format'] ) && ! empty( $feed['submission_date_format'] ) ) {
						$values_data[ $k ] = $this->wpssg_get_date( $entry[ $val[ $k ] ], $feed['submission_date_format'] );
					} else {
						$values_data[ $k ] = $this->wpssg_get_date( $entry[ $val[ $k ] ], $feed['submissiondate_format'] );
					}
				} elseif ( isset( $form_field_type[ $val[ $k ] ] ) && 'multiselect' === (string) $form_field_type[ $val[ $k ] ] ) {
					$wpssg_multiselect     = json_decode( $entry[ $val[ $k ] ] );
					$wpssg_multiselect_val = '';
					if ( null === $wpssg_multiselect ) { // check if it was invalid json string.
						$wpssg_multiselect_val = $entry[ $val[ $k ] ];
					}
					if ( is_array( $wpssg_multiselect ) ) {
						$wpssg_multiselect_val = implode( ',', $wpssg_multiselect );
					}
					$values_data[ $k ] = str_replace( array( '[', ']', '"', ',' ), array( '', '', '', ', ' ), self::$instance->wpssg_convert_to_readable( $wpssg_multiselect_val ) );
				} else {
					$values_data[ $k ] = self::$instance->wpssg_convert_to_readable( $entry[ $val[ $k ] ] );
				}
			}
		}

		return $values_data;
	}

	/**
	 * Return date.
	 *
	 * @param string $date Date.
	 * @param string $date_format Date Format.
	 */
	public function wpssg_get_date( $date, $date_format ) {

		if ( (bool) strtotime( $date ) ) {
			if ( isset( $date_format ) && 'withtime' === (string) $date_format ) {
				$date_format_filter = apply_filters( 'wpssg_date_format', self::$instance_api->wpssg_option( 'date_format' ) . ' ' . self::$instance_api->wpssg_option( 'time_format' ) );
				$date_val           = GFCommon::format_date( $date, false, $date_format_filter );
			} elseif ( isset( $date_format ) && 'withouttime' === (string) $date_format ) {
				$date_format_filter = apply_filters( 'wpssg_date_format', self::$instance_api->wpssg_option( 'date_format' ) );
				$date_val           = GFCommon::format_date( $date, false, $date_format_filter, false );

			} elseif ( isset( $date_format ) && 'default' === (string) $date_format ) {
				$date_format_filter = apply_filters( 'wpssg_date_format', self::$instance_api->wpssg_option( 'date_format' ) . ' ' . self::$instance_api->wpssg_option( 'time_format' ) );
				$date_val           = GFCommon::format_date( $date, false, $date_format_filter );
			} else {
				$date_val = get_date_from_gmt( $date, $date_format );
			}
		} else {
			$date_val = $date;
		}
		return $date_val;
	}

	/**
	 * Convert to Integer.
	 *
	 * @param string $data Entry ids array.
	 */
	public function wpssg_convert_int( $data ) {
		$data = array_map(
			function( $element ) {
				return ( is_numeric( $element ) ? (int) $element : $element );
			},
			$data
		);
		return $data;
	}

	/**
	 * Convert to String.
	 *
	 * @param string $data Entry ids array.
	 */
	public function wpssg_convert_string( $data ) {
		$data = array_map(
			function( $element ) {
				return ( is_string( $element ) ? (string) $element : $element );
			},
			$data
		);
		return $data;
	}
	/**
	 * Check spreadsheet and sheet exist or not.
	 *
	 * @param string $spreadsheetid Spreadsheet ID.
	 * @param string $sheetname Sheet Name.
	 */
	public function wpssg_check_sheet_exist( $spreadsheetid, $sheetname ) {
		$spreadsheets          = self::$instance->wpssg_list_googlespreedsheet();
		$spreadsheet_list      = array_column( $spreadsheets, 'value' );
		$data                  = array();
		$data['spreadsheetid'] = '';
		$data['sheetname']     = '';
		if ( '' !== $spreadsheetid && 'new' !== $spreadsheetid && in_array( $spreadsheetid, $spreadsheet_list, true ) ) {
			$data['spreadsheetid'] = $spreadsheetid;
			$response              = self::$instance_api->get_sheet_listing( $spreadsheetid );
			foreach ( $response->getSheets() as $s ) {
				$sheets[] = $s['properties']['title'];
			}
			if ( ! empty( $sheetname ) && in_array( $sheetname, $sheets, true ) ) {
				$data['sheetname'] = $sheetname;
			}
		}
		return $data;
	}
	/**
	 * Add additional map field values.
	 *
	 * @param array $fields map field values array.
	 * @return array $fields additional map field values array.
	 */
	public function wpssg_mapfield_choices( $fields ) {
		// phpcs:ignore
		$form_id = isset( $_GET['id'] ) ? sanitize_text_field( wp_unslash( $_GET['id'] ) ) : '';
		if ( ! empty( $form_id ) ) {
			$fileupload_fields = apply_filters( 'wpssg_fileupload_map_choices', array(), $form_id, $fields );
			if ( is_array( $fileupload_fields ) && ! empty( $fileupload_fields ) ) {
				$fields = $fileupload_fields;
			}
		}
		$is_mollie_active           = self::$instance->wpssg_is_mollie_active();
		$is_redsysforgravity_active = self::$instance->wpssg_is_redsysforgravity_active();

		$is_gpdfapi_active = self::$instance->wpssg_is_gpdfapi_active();

		if ( $is_mollie_active || $is_redsysforgravity_active ) {
			$fields['fields']['choices'][] = array(
				'label' => 'Payment Date',
				'value' => 'payment_date',
				'type'  => 'date',
			);
			$fields['fields']['choices'][] = array(
				'label' => 'Payment Amount',
				'value' => 'payment_amount',
				'type'  => 'text',
			);
			$fields['fields']['choices'][] = array(
				'label' => 'Payment Method',
				'value' => 'payment_method',
				'type'  => 'text',
			);
			$fields['fields']['choices'][] = array(
				'label' => 'Transaction ID',
				'value' => 'transaction_id',
				'type'  => 'text',
			);
		}
		if ( $is_gpdfapi_active ) {
			$fields['fields']['choices'][] = array(
				'label' => 'PDF',
				'value' => 'pdf',
				'type'  => 'text',
			);
		}

		$fields[] = array(
			'label'   => 'Visitor Information Fields',
			'choices' => array(
				array(
					'label' => esc_html__( 'User Agent', 'wpssg' ),
					'value' => 'user_agent',
				),
				array(
					'label' => esc_html__( 'IP Address', 'wpssg' ),
					'value' => 'ip_address',
				),
				array(
					'label' => esc_html__( 'Browser Name', 'wpssg' ),
					'value' => 'browser_name',
				),
				array(
					'label' => esc_html__( 'Operating System', 'wpssg' ),
					'value' => 'operating_system',
				),
				array(
					'label' => esc_html__( 'Device Type', 'wpssg' ),
					'value' => 'device_type',
				),
				array(
					'label' => esc_html__( 'City', 'wpssg' ),
					'value' => 'city',
				),
				array(
					'label' => esc_html__( 'State', 'wpssg' ),
					'value' => 'state',
				),
				array(
					'label' => esc_html__( 'Country', 'wpssg' ),
					'value' => 'country',
				),
				array(
					'label' => esc_html__( 'Country Code', 'wpssg' ),
					'value' => 'country_code',
				),
				array(
					'label' => esc_html__( 'Time Zone', 'wpssg' ),
					'value' => 'timezone',
				),
			),
		);
		return $fields;
	}
	/**
	 * Import Entries.
	 *
	 * @param int $wpssg_fid feed id.
	 * @param int $form_id form id.
	 */
	public function wpssg_import_entries( $wpssg_fid = '', $form_id = '' ) {

		$cron_function_call = false;
		if ( ! isset( $_POST['_wpnonce'] ) ) {
			$cron_function_call = true;
			if ( ! $wpssg_fid || ! $form_id ) {
				return;
			}
		} else {
			if ( ! isset( $_POST['_wpnonce'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['_wpnonce'] ) ), 'import_settings' ) ) {
				echo esc_html( 'nonce_error' );
				wp_die();
			}
			if ( isset( $_POST['fid'] ) && ! empty( $_POST['fid'] ) ) {
				$wpssg_fid = (int) sanitize_text_field( wp_unslash( $_POST['fid'] ) );
			} else {
				echo esc_html( 'error' );
				die;
			}
			if ( isset( $_POST['form_id'] ) && ! empty( $_POST['form_id'] ) ) {
				$form_id = sanitize_text_field( wp_unslash( $_POST['form_id'] ) );
			} else {
				echo esc_html( 'error' );
				die;
			}
		}
		$syncformfeed    = RGFormsModel::get_form_meta( $form_id );
		$form_field_type = self::$instance->wpssg_form_field_type( $syncformfeed );
		$fields_array    = self::$instance->wpssg_generate_headers( $syncformfeed );

		$wpssg_form_data = GFAPI::get_form( $form_id );

		$feeddata    = GFAPI::get_feed( $wpssg_fid, $form_id );
		$form_fields = isset( $wpssg_form_data['fields'] ) ? $wpssg_form_data['fields'] : array();

		$has_product_fields = array();
		$list_fields_array  = array();
		foreach ( $form_fields as $form_field ) {
			if ( GFCommon::is_product_field( $form_field['type'] ) ) {
				$has_product_fields[] = (int) $form_field['id'];
			} elseif ( 'list' === (string) $form_field['type'] && $form_field['enableColumns'] ) {
				$list_fields_array[] = $form_field['id'];
			}
		}
		$settings      = $feeddata['meta'] ? $feeddata['meta'] : array();
		$spreadsheetid = $settings['spreedsheets'] ? $settings['spreedsheets'] : '';
		$range         = $settings['sheetselect'] ? $settings['sheetselect'] : '';

		$sheet_data = self::$instance->wpssg_check_sheet_exist( $spreadsheetid, $range );
		if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
			if ( $cron_function_call ) {
				return;
			}
			echo esc_html( 'error' );
			die;
		}
		$formmapfieldsval = self::$instance->wpssg_get_mapfields( $spreadsheetid, $range );

		$allentry     = self::$instance_api->get_row_list( $spreadsheetid, $range );
		$entry_data   = $allentry->getValues();
		$data         = array();
		$entry_id_key = array_search( 'Entry Id', $formmapfieldsval, true );

		if ( false !== $entry_id_key ) {
			$data = array_map(
				function( $element ) {
					if ( isset( $element[ $entry_id_key ] ) ) {
						return $element[ $entry_id_key ];
					} else {
						return '';
					}
				},
				$data
			);
		}
		$data                   = self::$instance->wpssg_convert_int( $data );
		$field_ids              = array();
		$list_key               = '';
		$wpssg_prepared_headers = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $settings );
		$list_fields_arr        = array();
		foreach ( $formmapfieldsval as $fieldsval ) {
			if ( 'Entry Id' === (string) $fieldsval ) {
				continue;
			}
			$fieldname = str_replace( array( ' ', '.' ), '_', strtolower( trim( $fieldsval ) ) );
			if ( isset( $settings[ 'mappedFields_' . $fieldname ] ) ) {
				foreach ( $list_fields_array as $list_fieldskey ) {
					if ( 1 === strpos( '"' . $settings[ 'mappedFields_' . $fieldname ], $list_fieldskey . '.' ) ) {
						$list_fields_arr[ $list_fieldskey ][] = $settings[ 'mappedFields_' . $fieldname ];
						continue;
					}
				}
				if ( 'pdf' === (string) $settings[ 'mappedFields_' . $fieldname ] ) {
					unset( $settings[ 'mappedFields_' . $fieldname ] );
					continue;
				}
				$field_ids[ $settings[ 'mappedFields_' . $fieldname ] ] = $fieldsval;
				unset( $settings[ 'mappedFields_' . $fieldname ] );
			}
			if ( 'Submission Date' === (string) $fieldsval ) {
				$submissiondate = $settings['submissiondate'] ? $settings['submissiondate'] : 0;
				if ( 1 === (int) $submissiondate ) {
					$date_key                  = array_search( $fieldsval, $formmapfieldsval, true );
					$field_ids['date_created'] = $fieldsval;
				}
			}
			if ( 'Payment Status' === (string) $fieldsval ) {
				$paymentstatusheader = isset( $settings['paymentstatusheader'] ) ? $settings['paymentstatusheader'] : 0;
				if ( 1 === (int) $paymentstatusheader ) {
					$field_ids['payment_status'] = $fieldsval;
				}
			}
		}

		$total_entry_count = count( $entry_data ) - 1;
		if ( 1 === (int) count( $entry_data ) ) {
			if ( $cron_function_call ) {
				return;
			}
			echo esc_html( 'no_entries' );
			die;
		}
		if ( (int) $total_entry_count < 2000 ) {
			if ( (int) $total_entry_count > 50 ) {
				$entry_limit = 50;
			}
			if ( (int) $total_entry_count <= 50 ) {
				$entry_limit = (int) $total_entry_count;
			}
		} else {
			$entry_limit = 500;
		}
		$j           = 1;
		$entry_count = $entry_limit;
		$total_limit = $entry_limit;

		try {
			for ( $k = 0;$k < (float) ( $total_entry_count / $entry_limit );$k++ ) {
				set_time_limit( 999 );
				$entries     = array();
				$values_data = array();
				$entry_ids   = array();

				for ( $i = $j;$i <= $total_limit;$i++ ) {
					$field_values    = array();
					$field_values    = array(
						'form_id' => $form_id,
					);
					$list_col_values = array();

					foreach ( $field_ids as $field_id => $field_name ) {
						$key        = array_search( $field_name, $formmapfieldsval, true );
						$num        = $field_id;
						$form_field = GFFormsModel::get_field( $wpssg_form_data, $num );
						$is_explode = 0;
						if ( false !== strpos( $num, '.' ) ) {
							$fid        = explode( '.', $num );
							$num        = $fid[0];
							$is_explode = 1;
						}

						if ( ! isset( $entry_data[ $i ][ $key ] ) || ( isset( $entry_data[ $i ][ $key ] ) && empty( $entry_data[ $i ][ $key ] ) && false === array_key_exists( $num, $list_fields_arr ) ) || in_array( (int) $field_id, $has_product_fields, true ) ) {
							continue;
						}

						if ( array_key_exists( $num, $form_field_type ) ) {
							if ( 'checkbox' === (string) $form_field_type[ $num ] ) {
								$field                  = GFFormsModel::get_field( $wpssg_form_data, $num );
								$checkbox_value_arr     = array();
								$checkbox_value_arr     = explode( ',', $entry_data[ $i ][ $key ] );
								$checkbox_field_options = '';

								foreach ( $field['inputs'] as $choice ) {
									if ( in_array( $choice['label'], $checkbox_value_arr, true ) ) {
										$field_values[ $choice['id'] ] = $choice['label'];
									}
								}
								continue;
							} elseif ( 'list' === (string) $form_field_type[ $num ] && 0 === (int) $is_explode ) {
								$list_field        = GFFormsModel::get_field( $wpssg_form_data, $num );
								$is_enablecolumns  = isset( $list_field['enableColumns'] ) ? $list_field['enableColumns'] : 0;
								$list_value_string = '';
								$list_value_string = $entry_data[ $i ][ $key ];
								if ( 1 === (int) $is_enablecolumns ) {
									$list_columns = array();
									foreach ( $list_field['choices'] as $choice ) {
										$list_columns[] = isset( $choice['value'] ) ? $choice['value'] : '';
									}
									$list_column_count = count( $list_columns );
									if ( false !== strpos( $list_value_string, '|' ) ) {
										$list_rows_value = explode( '|', $list_value_string );
										$final_list      = array();
										foreach ( $list_rows_value as $list_row_value ) {
											$list_row = array();
											$list_row = explode( ',', trim( $list_row_value ) );
											for ( $li = 0;$li < $list_column_count;$li++ ) {
												if ( isset( $list_columns[ $li ] ) && isset( $list_row[ $li ] ) ) {
													$lists_value[ $list_columns[ $li ] ] = $list_row[ $li ];
												}
											}
											$final_list[] = $lists_value;
										}
										$field_values[ $num ] = maybe_serialize( $final_list );
									}
								} elseif ( ! empty( $list_value_string ) || '' !== (string) $list_value_string ) {
									$list_value           = explode( ',', $list_value_string );
									$field_values[ $num ] = maybe_serialize( $list_value );
								}
								continue;
							} elseif ( 'list' === (string) $form_field_type[ $num ] && 1 === (int) $is_explode ) {
								if ( false !== array_key_exists( $num, $list_fields_arr ) ) {
									$list_col_values[ $num ][ $field_id ] = isset( $entry_data[ $i ][ $key ] ) ? explode( ',', $entry_data[ $i ][ $key ] ) : array();
									$list_col_values_count                = count( $list_col_values[ $num ] );
									$list_fields_arr_count                = count( $list_fields_arr[ $num ] );

									if ( $list_col_values_count === $list_fields_arr_count ) {
										ksort( $list_col_values[ $num ] );
										$list_field   = GFFormsModel::get_field( $wpssg_form_data, $num );
										$list_columns = array();
										$li_col_count = 0;
										foreach ( $list_field['choices'] as $choice ) {
											$list_columns[ $num . '.' . $li_col_count ] = isset( $choice['value'] ) ? $choice['value'] : '';
											$li_col_count++;
										}
										$list_row_val       = array();
										$list_row_val_count = 0;
										foreach ( $list_col_values[ $num ] as $list_col_value_val ) {
											if ( count( $list_col_value_val ) > $list_row_val_count ) {
												$list_row_val_count = count( $list_col_value_val );
											}
										}
										foreach ( $list_col_values[ $num ] as $list_col_value_key => $list_col_value_val ) {
											$col_name = isset( $list_columns[ $list_col_value_key ] ) ? $list_columns[ $list_col_value_key ] : '';
											for ( $col_i = 0;$col_i < $list_row_val_count;$col_i++ ) {
												$list_row_val[ $col_i ][ $col_name ] = isset( $list_col_value_val[ $col_i ] ) ? $list_col_value_val[ $col_i ] : '';
											}
										}
										$field_values[ $num ] = maybe_serialize( $list_row_val );
										continue;
									}
									continue;
								}
							} elseif ( ( 'name' === (string) $form_field_type[ $num ] || 'address' === (string) $form_field_type[ $num ] ) && 0 === (int) $is_explode ) {

								$field            = GFFormsModel::get_field( $wpssg_form_data, $num );
								$name_value_arr   = array();
								$name_value_array = array();
								if ( false !== strpos( $entry_data[ $i ][ $key ], ',' ) ) {
									$name_value_arr = array_map( 'trim', explode( ',', trim( $entry_data[ $i ][ $key ] ) ) );
								} else {
									$name_value_arr = array_map( 'trim', explode( ' ', trim( $entry_data[ $i ][ $key ] ) ) );
								}
								$name_fields   = array();
								$enable_prefix = 0;

								foreach ( $field['inputs'] as $inputs ) {

									if ( ( isset( $inputs['isHidden'] ) && 1 !== (int) $inputs['isHidden'] ) || ! isset( $inputs['isHidden'] ) ) {

										if ( 'Prefix' === (string) $inputs['label'] ) {
											$enable_prefix = 1;
											$choices       = $inputs['choices'];
											$choice_values = array_column( $choices, 'value' );

											if ( in_array( trim( $name_value_arr[0] ), $choice_values, true ) ) {
												$field_values[ $inputs['id'] ] = $name_value_arr[0];
												unset( $name_value_arr[0] );
												$name_value_array = array_values( $name_value_arr );
												continue;
											}
										}

										$name_value_array             = array_values( $name_value_arr );
										$name_fields[ $inputs['id'] ] = $inputs['label'];
									}
								}
								ksort( $name_fields );
								$name_value_array_count = count( $name_value_array );

								for ( $n = 0;$n < $name_value_array_count;$n++ ) {
									$name_field_ids                        = array_keys( $name_fields );
									$field_values[ $name_field_ids[ $n ] ] = $name_value_array[ $n ];
								}
								continue;
							} elseif ( 'fileupload' === (string) $form_field_type[ $num ] && 1 === (int) $form_field['multipleFiles'] ) {
								$multiple_files_array = array();
								$multiple_files_array = explode( ',', $entry_data[ $i ][ $key ] );
								$multiple_files_array = array_filter( array_map( 'trim', $multiple_files_array ) );
								$field_values[ $num ] = wp_json_encode( $multiple_files_array );
								continue;
							}
						}

						if ( ( isset( $form_field_type[ $num ] ) && 'date' === (string) $form_field_type[ $num ] ) ) {

							$val       = strtotime( $entry_data[ $i ][ $key ] );
							$datefield = GFFormsModel::get_field( $wpssg_form_data, $num );
							// phpcs:ignore
							$dateformat = $datefield->dateFormat;
							switch ( (string) $dateformat ) {
								case 'mdy':
									$dateformat = 'm/d/Y';
									break;
								case 'dmy':
									$dateformat = 'd/m/Y';
									break;
								case 'dmy_dash':
									$dateformat = 'd-m-Y';
									break;
								case 'dmy_dot':
									$dateformat = 'd.m.Y';
									break;
								case 'ymd_slash':
									$dateformat = 'Y/m/d';
									break;
								case 'ymd_dash':
									$dateformat = 'Y-m-d';
									break;
								default:
											$dateformat = 'm/d/Y';
									break;
							}
							$date = gmdate( $dateformat, $val );

							$field_values[ $field_id ] = $date;
							continue;
						} elseif ( ( isset( $form_field_type[ $num ] ) && 'time' === (string) $form_field_type[ $num ] ) ) {
							$time = '';
							if ( ! empty( $entry_data[ $i ][ $key ] ) ) {
								$val       = '';
								$val       = strtotime( $entry_data[ $i ][ $key ] );
								$timefield = GFFormsModel::get_field( $wpssg_form_data, $num );
								if ( 12 === (int) $timefield['timeFormat'] ) {
									$time = gmdate( 'h:i a', $val );

								} else {
									$time = gmdate( 'H:i', $val );

								}
							}
							$field_values[ $field_id ] = $time;
							continue;
						}

						$key = array_search( $field_name, $formmapfieldsval, true );
						if ( 'Submission Date' === (string) $field_name || 'date_created' === (string) $field_id || 'payment_date' === (string) $field_id ) {
							if ( isset( $entry_data[ $i ][ $key ] ) ) {
								$val                       = strtotime( $entry_data[ $i ][ $key ] );
								$submission_datetime       = gmdate( 'Y-m-d H:i:s', $val );
								$field_values[ $field_id ] = $submission_datetime;
								continue;
							}
						}
						if ( 'form_title' === (string) $field_id || 'id' === (string) $field_id ) {
							continue;
						}
						if ( 'ip_address' === (string) $field_id || 'ip' === (string) $field_id ) {
							$field_values['ip'] = isset( $entry_data[ $i ][ $key ] ) ? $entry_data[ $i ][ $key ] : '';
							continue;
						}
						if ( 'payment_amount' === (string) $field_id ) {
							$field_values['payment_amount'] = isset( $entry_data[ $i ][ $key ] ) ? GFCommon::to_number( $entry_data[ $i ][ $key ] ) : '';
							continue;
						}
						$field_values[ $field_id ] = isset( $entry_data[ $i ][ $key ] ) ? $entry_data[ $i ][ $key ] : '';
					}
					$entry_id = GFAPI::add_entry( $field_values, $form_id );
					$entry    = GFAPI::get_entry( $entry_id );

					$entry_ids[] = $entry_id;
					if ( GFCommon::has_post_field( $form_fields ) ) {
						$entry['post_id']  = GFFormsModel::create_post( $wpssg_form_data, $entry );
						$entry_updateddata = GFAPI::update_entry( $entry );
					}
				}

				foreach ( $entry_ids as $entryid ) {
					$entry = GFAPI::get_entry( $entryid );
					if ( ! is_wp_error( $entry ) ) {
						$values_data[] = self::$instance->wpssg_prepared_values( $wpssg_form_data, $wpssg_prepared_headers, $entry, $settings );
					}
				}

				if ( is_array( $values_data ) && ! empty( $values_data ) ) {
					try {
						$startindex = $j + 1;
						/*Update entry*/
						$rangetoupdate = $range . '!A' . $startindex;

						$requestbody = self::$instance_api->valuerangeobject( $values_data );
						$params      = self::$instance->get_row_format( $settings );
						$param       = self::$instance_api->setparamater( $spreadsheetid, $rangetoupdate, $requestbody, $params );
						$response    = self::$instance_api->updateentry( $param );

					} catch ( Exception $e ) {
						if ( ! $cron_function_call ) {
							$allowed_html = wp_kses_allowed_html( 'post' );
							echo wp_kses( 'Message : ' . $e->getMessage(), $allowed_html );
						}
					}
				}
				$j          += $entry_limit;
				$total_limit = ( $entry_count + $entry_limit ) < $total_entry_count ? ( $entry_count + $entry_limit ) : $total_entry_count;
				$entry_count = ( $entry_count + $entry_limit ) < $total_entry_count ? ( $entry_count + $entry_limit ) : ( $total_entry_count - $entry_limit );
			}
			if ( $cron_function_call ) {
				return;
			}
			if ( ! is_wp_error( $entry_ids ) ) {
				echo esc_html( 'successful' );
				die;
			} else {
				echo esc_html( 'error' );
				die;
			}
		} catch ( Exception $e ) {
			if ( $cron_function_call ) {
				return;
			}
			echo esc_html( 'error' );
			die;
		}
	}
	/**
	 * Check feed is saved or not.
	 *
	 * @access public
	 */
	public function wpssg_check_feed() {
		// @codingStandardsIgnoreStart.
		if ( isset( $_GET['id'] ) && isset( $_GET['fid'] ) && 0 !== (int) sanitize_text_field( wp_unslash( $_GET['fid'] ) ) ) {
			$wpssg_fid = sanitize_text_field( wp_unslash( $_GET['fid'] ) );
			$form_id   = sanitize_text_field( wp_unslash( $_GET['id'] ) );
			// @codingStandardsIgnoreEnd.
			$feeddata = GFAPI::get_feed( $wpssg_fid, $form_id );

			$settings = $feeddata['meta'] ? $feeddata['meta'] : array();

			if ( isset( $settings['spreedsheets'] ) && isset( $settings['sheetselect'] ) ) {
				$sheet_data = self::$instance->wpssg_check_sheet_exist( $settings['spreedsheets'], $settings['sheetselect'] );
				if ( '' !== (string) $sheet_data['spreadsheetid'] && '' !== (string) $sheet_data['sheetname'] ) {
					return true;
				}
			}
			return false;
		} else {
			return false;
		}
	}
	/**
	 * Check feed settings are saved or not.
	 *
	 * @access public
	 */
	public function wpssg_check_feedsetting() {
		// @codingStandardsIgnoreStart.
		if ( isset( $_GET['id'] ) && isset( $_GET['fid'] ) && 0 !== (int) sanitize_text_field( wp_unslash( $_GET['fid'] ) ) ) {
			$wpssg_fid = sanitize_text_field( wp_unslash( $_GET['fid'] ) );
			$form_id   = sanitize_text_field( wp_unslash( $_GET['id'] ) );
			// @codingStandardsIgnoreEnd.
			$feeddata = GFAPI::get_feed( $wpssg_fid, $form_id );

			$settings = $feeddata['meta'] ? $feeddata['meta'] : array();

			if ( isset( $settings['spreedsheets'] ) && isset( $settings['sheetselect'] ) ) {
				$sheet_data = self::$instance->wpssg_check_sheet_exist( $settings['spreedsheets'], $settings['sheetselect'] );
				if ( '' !== (string) $sheet_data['spreadsheetid'] && '' !== (string) $sheet_data['sheetname'] && isset( $settings['row_format'] ) ) {
					return true;
				}
			}
			return false;
		} else {
			return false;
		}
	}
	/**
	 * Process the single action (trash and restore) of entry.
	 *
	 * @param int    $entry_id The entry id currently being processed.
	 * @param string $property_value The current updated status of entry.
	 * @param string $previous_value The previous status of entry.
	 * @param int    $form_id The form id currently being processed.
	 */
	public function wpssg_delete_entry_post( $entry_id, $property_value, $previous_value, $form_id = '' ) {
		// @codingStandardsIgnoreStart.
		if ( ( isset( $_POST['action'] ) && ( 'trash' === sanitize_text_field( wp_unslash( $_POST['action'] ) ) || 'restore' === sanitize_text_field( wp_unslash( $_POST['action'] ) ) ) ) && ( ( isset( $_POST['paged'] ) && sanitize_text_field( wp_unslash( $_POST['paged'] ) ) > 0 ) || ( isset( $_POST['entry'] ) && is_array( $_POST['entry'] ) ) ) ) {
			return;
		}
		
		// @codingStandardsIgnoreEnd.
		if ( ! $entry_id || ! self::$instance_api->checkcredenatials() ) {
			return;
		}
		$entry = GFAPI::get_entry( $entry_id );
		if ( ! $form_id ) {
			$form_id = $entry['form_id'];
		}

		if ( 'trash' === (string) $property_value || ( 'trash' === (string) $previous_value && 'active' === (string) $property_value ) ) {

			$formfeed              = self::$instance->get_feeds( $form_id );
			$formfeedcountdata     = ! is_wp_error( $formfeed ) ? count( $formfeed ) : 0;
			$wpssg_form_data       = GFAPI::get_form( $form_id );
			$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );
			$wpssg_data            = array();
			try {
				for ( $i = 0; $i < $formfeedcountdata; $i++ ) {
					if ( '1' === (string) $formfeed[ $i ]['is_active'] ) {
						$feed = $formfeed[ $i ]['meta'];
						if ( $is_creditcard_enabled && ! is_wp_error( $entry ) ) {
							if ( isset( $feed['paymentstatus'] ) && $feed['paymentstatus'] ) {
								if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
									continue;
								}
							}
						}
						if ( isset( $feed['spreedsheets'] ) && isset( $feed['sheetselect'] ) ) {
							$sheet_data = self::$instance->wpssg_check_sheet_exist( $feed['spreedsheets'], $feed['sheetselect'] );
							if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
								continue;
							}
							$formmapfieldsval    = self::$instance->wpssg_get_mapfields( $feed['spreedsheets'], $feed['sheetselect'] );
							$val                 = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
							$wpssg_spreadsheetid = $feed['spreedsheets'];
							$wpssg_sheetname     = $feed['sheetselect'];
							if ( is_array( $formmapfieldsval ) ) {
								$wpssg_entryid_key = array_search( 'Entry Id', $formmapfieldsval, true );
								$wpssg_alphaindex  = '';
								if ( false !== $wpssg_entryid_key ) {

									$wpssg_alphaindex = self::$instance->wpssg_get_column_index( $wpssg_entryid_key + 1 );
								} elseif ( in_array( 'id', $val, true ) ) {
									$wpssg_entryid_key = array_search( 'id', $val, true );
									$wpssg_alphaindex  = self::$instance->wpssg_get_column_index( $wpssg_entryid_key + 1 );
								}
								if ( '' !== (string) $wpssg_alphaindex ) {
									$wpssg_rangetofind = $wpssg_sheetname . '!' . $wpssg_alphaindex . ':' . $wpssg_alphaindex;
									$wpssg_allentry    = self::$instance_api->get_row_list( $wpssg_spreadsheetid, $wpssg_rangetofind );
									$wpssg_data        = $wpssg_allentry->getValues();

									$wpssg_data = array_map(
										function( $wpssg_element ) {
											if ( isset( $wpssg_element['0'] ) ) {
												return $wpssg_element['0'];
											} else {
												return '';
											}
										},
										$wpssg_data
									);
								}
								$wpssg_num           = array_search( (int) $entry_id, self::$instance->wpssg_convert_int( $wpssg_data ), true );
								$response            = self::$instance_api->get_sheet_listing( $wpssg_spreadsheetid );
								$existingsheetsnames = self::$instance_api->get_sheet_list( $response );

								$sheetid = $existingsheetsnames[ $wpssg_sheetname ];
								if ( 'trash' === (string) $property_value ) {
									if ( $wpssg_num > 0 ) {

										$param               = array();
										$param['sheetid']    = $sheetid;
										$param['startindex'] = $wpssg_num;
										$param['endindex']   = $wpssg_num + 1;

										$deleterequest          = self::$instance_api->deleteDimensionrequests( $param, 'ROWS' );
										$param                  = array();
										$param['spreadsheetid'] = $wpssg_spreadsheetid;
										$param['requestarray']  = $deleterequest;
										self::$instance_api->updatebachrequests( $param );

									}
								}
								if ( 'trash' === (string) $previous_value && 'active' === (string) $property_value ) {
									$form = GFAPI::get_form( $form_id );

									$is_condition_enabled = rgar( $feed, 'feed_condition_conditional_logic' ) === '1';
									$logic                = rgars( $feed, 'feed_condition_conditional_logic_object/conditionalLogic' );
									if ( $is_condition_enabled && ! empty( $logic ) ) {
										$check = GFCommon::evaluate_conditional_logic( $logic, $form, $entry );
										if ( ! $check ) {
											continue;
										}
									}
									$val         = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
									$values_data = self::$instance->wpssg_prepared_values( $form, $val, $entry, $feed );

									$values = array( $values_data );

									$requestbody = self::$instance_api->valuerangeobject( $values );

									$params = self::$instance->get_row_format( $feed );

									if ( $wpssg_num > 0 ) {
											$wpssg_num++;
											$range    = $wpssg_sheetname . '!A' . $wpssg_num;
											$param    = self::$instance_api->setparamater( $wpssg_spreadsheetid, $range, $requestbody, $params );
											$response = self::$instance_api->updateentry( $param );
									} else {
										$wpssg_append = 0;
										foreach ( $wpssg_data as $wpssg_key => $wpssg_value ) {
											if ( ! empty( $wpssg_value ) ) {
												if ( ( (int) $entry_id < (int) $wpssg_value ) ) {

													$wpssg_append = 1;

													$startindex = $wpssg_key;
													$endindex   = $wpssg_key + 1;

													$param               = array();
													$param['sheetid']    = $sheetid;
													$param['startindex'] = $startindex;
													$param['endindex']   = $endindex;

													$batchupdaterequest = self::$instance_api->insertdimensionobject( $param );

													$requestobject                  = array();
													$requestobject['spreadsheetid'] = $wpssg_spreadsheetid;
													$requestobject['requestbody']   = $batchupdaterequest;
													$response                       = self::$instance_api->formatsheet( $requestobject );
													$startindex++;
													$range    = $wpssg_sheetname . '!A' . $startindex;
													$param    = self::$instance_api->setparamater( $wpssg_spreadsheetid, $range, $requestbody, $params );
													$response = self::$instance_api->updateentry( $param );
													break;

												}
											}
										}
										$wpssg_rangetofind = $wpssg_sheetname . '!A:Z';
										$wpssg_allentry    = self::$instance_api->get_row_list( $wpssg_spreadsheetid, $wpssg_rangetofind );
										$wpssg_data        = $wpssg_allentry->getValues();

										$wpssg_data = array_map(
											function( $wpssg_element ) {
												if ( isset( $wpssg_element['0'] ) ) {
													return $wpssg_element['0'];
												} else {
													return '';
												}
											},
											$wpssg_data
										);
										if ( 0 === (int) $wpssg_append ) {
											$wpssg_isupdated = 0;

											$requestbody = self::$instance_api->valuerangeobject( $values );
											$params      = self::$instance->get_row_format( $feed );

											if ( count( $wpssg_data ) > 1 ) {

												$range = $wpssg_sheetname . '!A' . ( count( $wpssg_data ) + 1 );

												$param    = self::$instance_api->setparamater( $wpssg_spreadsheetid, $range, $requestbody, $params );
												$response = self::$instance_api->updateentry( $param );

												$wpssg_isupdated = 1;
											}
											if ( 0 === (int) $wpssg_isupdated ) {
												$param    = self::$instance_api->setparamater( $wpssg_spreadsheetid, $range, $requestbody, $params );
												$response = self::$instance_api->appendentry( $param );
											}
										}
									}
								}
							}
						}
					}
				}
			} catch ( Exception $e ) {
				$allowed_html = wp_kses_allowed_html( 'post' );
				echo wp_kses( 'Message : ' . $e->getMessage(), $allowed_html );
			}
		}
	}
	/**
	 * Process the delete operation.
	 *
	 * @param int   $entry_id The entry id currently being processed.
	 * @param array $entry The entry currently being processed.
	 */
	public function wpssg_delete_view_entry( $entry_id, $entry = array() ) {
		// phpcs:ignore
		if ( isset( $_GET['filter'] ) && 'trash' === sanitize_text_field( wp_unslash( $_GET['filter'] ) ) ) {
			return;
		}
		if ( ! $entry ) {
			$entry = GFAPI::get_entry( $entry_id );
		}
		self::wpssg_delete_entry_post( $entry_id, 'trash', $entry['status'], $entry['form_id'] );
	}
	/**
	 * Process the bulk action(Trash and Restore) perform on entry list.
	 *
	 * @param string $action The action to be processed.
	 * @param array  $entry_ids The entry ids currently being processed.
	 * @param int    $form_id The form id currently being processed.
	 */
	public function wpssg_entry_list_bulk_action( $action, $entry_ids, $form_id ) {
		// phpcs:ignore
		if ( isset( $_GET['filter'] ) && 'trash' === sanitize_text_field( wp_unslash( $_GET['filter'] ) ) && 'restore' !== (string) $action ) {
			return;
		}
		if ( 'trash' === (string) $action || 'restore' === (string) $action ) {

			if ( empty( $entry_ids ) || ! $form_id ) {
				return;
			}
			$formfeed              = self::$instance->get_feeds( $form_id );
			$formfeedcountdata     = ! is_wp_error( $formfeed ) ? count( $formfeed ) : 0;
			$wpssg_form_data       = GFAPI::get_form( $form_id );
			$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );
			try {
				for ( $i = 0; $i < $formfeedcountdata; $i++ ) {
					if ( '1' === (string) $formfeed[ $i ]['is_active'] ) {
						$feed = $formfeed[ $i ]['meta'];

						if ( isset( $feed['spreedsheets'] ) && isset( $feed['sheetselect'] ) ) {
							$sheet_data = self::$instance->wpssg_check_sheet_exist( $feed['spreedsheets'], $feed['sheetselect'] );
							if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
								continue;
							}

							$formmapfieldsval    = self::$instance->wpssg_get_mapfields( $feed['spreedsheets'], $feed['sheetselect'] );
							$val                 = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
							$wpssg_spreadsheetid = $feed['spreedsheets'];
							$wpssg_sheetname     = $feed['sheetselect'];

							$wpssg_data    = array();
							$deleterequest = array();
							if ( is_array( $formmapfieldsval ) ) {
								$wpssg_entryid_key = array_search( 'Entry Id', $formmapfieldsval, true );
								$wpssg_alphaindex  = '';
								if ( false !== $wpssg_entryid_key ) {

									$wpssg_alphaindex = self::$instance->wpssg_get_column_index( $wpssg_entryid_key + 1 );
								} elseif ( in_array( 'id', $val, true ) ) {
									$wpssg_entryid_key = array_search( 'id', $val, true );
									$wpssg_alphaindex  = self::$instance->wpssg_get_column_index( $wpssg_entryid_key + 1 );
								}
								if ( '' !== (string) $wpssg_alphaindex ) {
									$wpssg_rangetofind = $wpssg_sheetname . '!' . $wpssg_alphaindex . ':' . $wpssg_alphaindex;
									$wpssg_allentry    = self::$instance_api->get_row_list( $wpssg_spreadsheetid, $wpssg_rangetofind );
									$wpssg_data        = $wpssg_allentry->getValues();

									$wpssg_data = array_map(
										function( $wpssg_element ) {
											if ( isset( $wpssg_element['0'] ) ) {
												return $wpssg_element['0'];
											} else {
												return '';
											}
										},
										$wpssg_data
									);

								}
								$response            = self::$instance_api->get_sheet_listing( $wpssg_spreadsheetid );
								$existingsheetsnames = self::$instance_api->get_sheet_list( $response );
								$sheetid             = $existingsheetsnames[ $wpssg_sheetname ];

								foreach ( $entry_ids as $entry_id ) {
									if ( $is_creditcard_enabled ) {
										$entry = GFAPI::get_entry( $entry_id );
										if ( isset( $feed['paymentstatus'] ) && $feed['paymentstatus'] ) {
											if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
												continue;
											}
										}
									}
									$wpssg_num = '';
									$wpssg_num = array_search( (int) $entry_id, self::$instance->wpssg_convert_int( $wpssg_data ), true );

									if ( 'trash' === (string) $action ) {
										// Delete all trashed entries from the selected sheet.
										if ( $wpssg_num > 0 ) {
											$param               = array();
											$param['sheetid']    = $sheetid;
											$param['startindex'] = $wpssg_num;
											$param['endindex']   = $wpssg_num + 1;

											$deleterequest[] = self::$instance_api->deleteDimensionrequests( $param, 'ROWS' );
										}
									}
								}
								if ( ! empty( $deleterequest ) ) {
									global $keys;
									$keys = array();
									array_walk_recursive(
										$deleterequest,
										function( $value, $key ) {
											global $keys;
											if ( 'startIndex' === (string) $key ) {
												$keys[] = $value;
											}
										}
									);

									if ( count( $keys ) === count( $deleterequest ) ) {
										array_multisort( $keys, SORT_DESC, $deleterequest );
									}

									$param                  = array();
									$param['spreadsheetid'] = $wpssg_spreadsheetid;
									$param['requestarray']  = $deleterequest;
									$response               = self::$instance_api->updatebachrequests( $param );
								}
								if ( 'restore' === (string) $action ) {
									// Append all restored entries to last row of the selected sheet.
									$values_data      = array();
									$total_entries    = count( $entry_ids );
									$val              = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
									$form             = GFAPI::get_form( $form_id );
									$sorted_entry_ids = $entry_ids;
									sort( $sorted_entry_ids );
									for ( $j = 0; $j < $total_entries; $j++ ) {
										$entry = array();
										$entry = GFAPI::get_entry( $sorted_entry_ids[ $j ] );
										if ( $is_creditcard_enabled ) {
											if ( isset( $feed['paymentstatus'] ) && $feed['paymentstatus'] ) {
												if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
													continue;
												}
											}
										}
										$is_condition_enabled = rgar( $feed, 'feed_condition_conditional_logic' ) === '1';
										$logic                = rgars( $feed, 'feed_condition_conditional_logic_object/conditionalLogic' );
										if ( $is_condition_enabled && ! empty( $logic ) ) {
											$check = GFCommon::evaluate_conditional_logic( $logic, $form, $entry );
											if ( ! $check ) {
												continue;
											}
										}

										$values_data[] = self::$instance->wpssg_prepared_values( $form, $val, $entry, $feed );
									}
									if ( is_array( $values_data ) && ! empty( $values_data ) ) {
										$startindex = count( $wpssg_data ) + 1;
										$endindex   = count( $entry_ids ) + $startindex;

										$param               = array();
										$param['sheetid']    = $sheetid;
										$param['startindex'] = $startindex;
										$param['endindex']   = $endindex;

										$batchupdaterequest = self::$instance_api->insertdimensionobject( $param );

										$requestobject                  = array();
										$requestobject['spreadsheetid'] = $wpssg_spreadsheetid;
										$requestobject['requestbody']   = $batchupdaterequest;
										$response                       = self::$instance_api->formatsheet( $requestobject );

										/*Append entry*/
										$rangetoupdate = $wpssg_sheetname . '!A' . $startindex;

										$requestbody = self::$instance_api->valuerangeobject( $values_data );
										$params      = self::$instance->get_row_format( $feed );
										$param       = self::$instance_api->setparamater( $wpssg_spreadsheetid, $rangetoupdate, $requestbody, $params );
										$response    = self::$instance_api->updateentry( $param );
									}
								}
							}
						}
					}
				}
			} catch ( Exception $e ) {
				$allowed_html = wp_kses_allowed_html( 'post' );
				echo wp_kses( 'Message : ' . $e->getMessage(), $allowed_html );
			}
		}
	}
	/**
	 * Process the update payment status action for entry.
	 *
	 * @param int $entry_id The entry id currently being processed.
	 */
	public function wpssg_update_payment_status( $entry_id ) {

		if ( ! $entry_id || ! self::$instance_api->checkcredenatials() ) {
			return;
		}
		$entry = GFAPI::get_entry( $entry_id );

		$form_id = $entry['form_id'];
		// Get the plugin settings.
		$wpssg_google_settings = self::$instance_api->wpssg_option( 'wpssg_google_settings' );
		// Access a specific setting e.g. an authToken.
		if ( isset( $wpssg_google_settings[2] ) && ! empty( $wpssg_google_settings[2] ) ) {

			$form_field_type       = array();
			$form                  = GFAPI::get_form( $form_id );
			$formfeed              = self::$instance->get_feeds( $form_id );
			$form_field_type       = self::$instance->wpssg_form_field_type( $form );
			$countformfeed         = ! is_wp_error( $formfeed ) ? count( $formfeed ) : 0;
			$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $form );

			for ( $i = 0;$i < $countformfeed; $i++ ) {
				$is_update_entry = 0;
				if ( 1 === (int) $formfeed[ $i ]['is_active'] ) {
					if ( $is_creditcard_enabled ) {
						if ( isset( $formfeed[ $i ]['meta']['paymentstatus'] ) && $formfeed[ $i ]['meta']['paymentstatus'] ) {
							if ( ! in_array( $entry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
								continue;
							}
						}
					}
					$feed = $formfeed[ $i ]['meta'];

					$flag                 = 1;
					$is_condition_enabled = rgar( $feed, 'feed_condition_conditional_logic' ) === '1';
					$logic                = rgars( $feed, 'feed_condition_conditional_logic_object/conditionalLogic' );

					if ( $is_condition_enabled && ! empty( $logic ) ) {
						$check = GFCommon::evaluate_conditional_logic( $logic, $form, $entry );
						if ( ! $check ) {
							continue;
						}
					}
					if ( ! isset( $feed['spreedsheets'] ) || empty( $feed['spreedsheets'] ) || ! isset( $feed['sheetselect'] ) || empty( $feed['sheetselect'] ) ) {
						continue;
					}
					$spreadsheetid = $feed['spreedsheets'];
					$range         = $feed['sheetselect'];

					$sheet_data = self::$instance->wpssg_check_sheet_exist( $feed['spreedsheets'], $feed['sheetselect'] );
					if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
						continue;
					}
					$response                  = self::$instance_api->get_sheet_listing( $spreadsheetid );
					$wpssg_existingsheetsnames = self::$instance_api->get_sheet_list( $response );

					$sheetid = $wpssg_existingsheetsnames[ $range ];

					$allentry = self::$instance_api->get_row_list( $spreadsheetid, $range . '!A:Z' );
					$data     = $allentry->getValues();

					$data = array_map(
						function( $element ) {
							if ( isset( $element['0'] ) ) {
								return $element['0'];
							} else {
								return '';
							}
						},
						$data
					);

					$values_data      = array();
					$formmapfieldsval = self::$instance->wpssg_get_mapfields( $spreadsheetid, $range );

					$countformmapfieldsval = count( $formmapfieldsval );
					$val                   = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $feed );
					$values_data           = self::$instance->wpssg_prepared_values( $form, $val, $entry, $feed );

					$values = array( $values_data );

					$requestbody = self::$instance_api->valuerangeobject( $values );

					$params = self::$instance->get_row_format( $feed );
					$data   = self::$instance->wpssg_convert_int( $data );

					$num = array_search( (int) $entry_id, $data, true );

					if ( $num > 0 ) {
							$num++;
							$range    = $range . '!A' . $num;
							$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
							$response = self::$instance_api->updateentry( $param );

					} else {

						$wpssg_append = 0;
						foreach ( $data as $wpssg_key => $wpssg_value ) {
							if ( ! empty( $wpssg_value ) ) {
								if ( ( (int) $entry_id < (int) $wpssg_value ) ) {

									$wpssg_append = 1;

									$startindex = $wpssg_key;
									$endindex   = $wpssg_key + 1;

									$param               = array();
									$param['sheetid']    = $sheetid;
									$param['startindex'] = $startindex;
									$param['endindex']   = $endindex;

									$batchupdaterequest = self::$instance_api->insertdimensionobject( $param );

									$requestobject                  = array();
									$requestobject['spreadsheetid'] = $spreadsheetid;
									$requestobject['requestbody']   = $batchupdaterequest;
									$response                       = self::$instance_api->formatsheet( $requestobject );
									$startindex++;
									$range    = $range . '!A' . $startindex;
									$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
									$response = self::$instance_api->updateentry( $param );
									break;

								}
							}
						}
						if ( 0 === (int) $wpssg_append ) {
							$wpssg_isupdated = 0;

							$requestbody = self::$instance_api->valuerangeobject( $values );
							$params      = self::$instance->get_row_format( $feed );

							if ( count( $data ) > 1 ) {

								$range = $range . '!A' . ( count( $data ) + 1 );

								$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
								$response = self::$instance_api->updateentry( $param );

								$wpssg_isupdated = 1;
							}
							if ( 0 === (int) $wpssg_isupdated ) {
								$param    = self::$instance_api->setparamater( $spreadsheetid, $range, $requestbody, $params );
								$response = self::$instance_api->appendentry( $param );
							}
						}
					}
				}
			}
		} else {
			return;
		}

	}

	/** CRON FUNCTION START **/
	/**
	 * Add cron interval.
	 *
	 * @param array $schedules .
	 */
	public static function wpssg_add_cron_interval( $schedules ) {
		$wpssg_interval              = 86400 * 14;
		$schedules['thirty_minutes'] = array(
			'interval' => 1800,
			'display'  => esc_html__( 'Every Thirty Minutes', 'wpssg' ),
		);
		$schedules['fifteen_days']   = array(
			'interval' => $wpssg_interval,
			'display'  => esc_html__( 'Every Fifteen Days', 'wpssg' ),
		);
		return $schedules;
	}
	/**
	 * Update sync schedule data.
	 *
	 * @param array $data post data.
	 * @param int   $fid feed id.
	 * @param array $settings feed settings.
	 */
	public static function wpssg_update_schedule_data( $data, $fid, $settings ) {
		if ( ! isset( $_POST['wpssg_general_settings'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['wpssg_general_settings'] ) ), 'save_general_settings' ) ) {
			echo esc_html__( 'Sorry, your nonce did not verify.', 'wpssg' );
			die();
		}
		$wpssg_schedule_enable               = isset( $_POST['scheduling_enable'] ) ? sanitize_text_field( wp_unslash( $_POST['scheduling_enable'] ) ) : '';
		$settings['wpssg_scheduling_enable'] = $wpssg_schedule_enable;
		$form_id                             = isset( $_POST['form_id'] ) ? sanitize_text_field( wp_unslash( $_POST['form_id'] ) ) : '';
		$form_args                           = array(
			'feed_id' => $fid,
			'form_id' => $form_id,
		);

		if ( 1 === (int) $wpssg_schedule_enable ) {

			$wpssg_scheduling_run_on = isset( $settings['wpssg_scheduling_run_on'] ) ? $settings['wpssg_scheduling_run_on'] : '';
			$scheduling_run_on       = isset( $_POST['scheduling_run_on'] ) ? sanitize_text_field( wp_unslash( $_POST['scheduling_run_on'] ) ) : '';
			if ( (string) $wpssg_scheduling_run_on !== (string) $scheduling_run_on ) {
				self::$instance->wpssg_remove_cron_job( $form_args );
			}
			$settings['wpssg_scheduling_run_on'] = $scheduling_run_on;
			if ( 'recurrence' === (string) $scheduling_run_on ) {
				$wpssg_interval                        = array(
					'thirty_minutes' => 'thirty_minutes',
					'once_daily'     => 'daily',
					'twice_daily'    => 'twicedaily',
					'fifteen_days'   => 'fifteen_days',
					'monthly'        => 'monthly',
				);
				$schedule_recurrence                   = isset( $_POST['schedule_recurrence'] ) ? sanitize_text_field( wp_unslash( $_POST['schedule_recurrence'] ) ) : '';
				$wpssg_schedule_recurrence             = isset( $settings['wpssg_schedule_recurrence'] ) ? $settings['wpssg_schedule_recurrence'] : '';
				$settings['wpssg_schedule_recurrence'] = $schedule_recurrence;
				if ( isset( $wpssg_interval[ $schedule_recurrence ] ) ) {
					if ( (string) $schedule_recurrence !== (string) $wpssg_schedule_recurrence ) {
						self::$instance->wpssg_remove_cron_job( $form_args );
					}
					self::$instance->wpssg_set_cron_job( time(), $wpssg_interval[ $schedule_recurrence ], $form_args );
				}
			}
			if ( 'weekly' === (string) $scheduling_run_on ) {
				$scheduling_weekly_days       = isset( $_POST['scheduling_weekly_days'] ) ? sanitize_text_field( wp_unslash( $_POST['scheduling_weekly_days'] ) ) : '';
				$wpssg_scheduling_weekly_days = isset( $settings['wpssg_scheduling_weekly_days'] ) ? $settings['wpssg_scheduling_weekly_days'] : '';
				if ( ! empty( $scheduling_weekly_days ) ) {
					self::$instance->wpssg_remove_cron_job( $form_args );
					$wpssg_weekdays = explode( ',', $scheduling_weekly_days );
					if ( is_array( $wpssg_weekdays ) && ! empty( $wpssg_weekdays ) ) {
						$settings['wpssg_scheduling_weekly_days'] = $scheduling_weekly_days;
						foreach ( $wpssg_weekdays as $weekday ) {
							$startcron = strtotime( 'next ' . $weekday );
							self::$instance->wpssg_set_cron_job( $startcron, 'weekly', $form_args );
						}
					}
				}
			}
			if ( 'onetime' === (string) $scheduling_run_on ) {
				$scheduling_date                   = isset( $_POST['scheduling_date'] ) ? sanitize_text_field( wp_unslash( $_POST['scheduling_date'] ) ) : '';
				$scheduling_time                   = isset( $_POST['scheduling_time'] ) ? sanitize_text_field( wp_unslash( $_POST['scheduling_time'] ) ) : '';
				$wpssg_scheduling_date             = isset( $settings['wpssg_scheduling_date'] ) ? $settings['wpssg_scheduling_date'] : '';
				$wpssg_scheduling_time             = isset( $settings['wpssg_scheduling_time'] ) ? $settings['wpssg_scheduling_time'] : '';
				$settings['wpssg_scheduling_date'] = $scheduling_date;
				$settings['wpssg_scheduling_time'] = $scheduling_time;
				if ( ! empty( $scheduling_date ) && ! empty( $scheduling_time ) ) {
					$starttime = strtotime( $scheduling_date . ' ' . $scheduling_time );
					if ( ! empty( $wpssg_scheduling_date ) && ! empty( $wpssg_scheduling_time ) ) {
						$previousstarttime = strtotime( $wpssg_scheduling_date . ' ' . $wpssg_scheduling_time );
						if ( (int) $starttime !== (int) $previousstarttime ) {
							self::$instance->wpssg_remove_cron_job( $form_args );
						}
					}
					self::$instance->wpssg_set_cron_job( $starttime, '', $form_args, true );
				}
			}
		} else {
			self::$instance->wpssg_remove_cron_job( $form_args );
		}
		$result = self::$instance->update_feed_meta( $fid, $settings );
		return $settings;
	}
	/**
	 * Update import schedule data.
	 *
	 * @param array $data .
	 * @param int   $fid feed id.
	 * @param array $settings feed settings.
	 */
	public static function wpssg_update_import_schedule_data( $data, $fid, $settings ) {
		if ( ! isset( $_POST['wpssg_general_settings'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_POST['wpssg_general_settings'] ) ), 'save_general_settings' ) ) {
			echo esc_html__( 'Sorry, your nonce did not verify.', 'wpssg' );
			die();
		}
		$wpssg_import_schedule_enable               = isset( $_POST['import_scheduling_enable'] ) ? sanitize_text_field( wp_unslash( $_POST['import_scheduling_enable'] ) ) : '';
		$settings['wpssg_import_scheduling_enable'] = $wpssg_import_schedule_enable;

		$form_id   = isset( $_POST['form_id'] ) ? sanitize_text_field( wp_unslash( $_POST['form_id'] ) ) : '';
		$form_args = array(
			'feed_id' => $fid,
			'form_id' => $form_id,
		);

		if ( 1 === (int) $wpssg_import_schedule_enable ) {

			$wpssg_import_scheduling_run_on = isset( $settings['wpssg_import_scheduling_run_on'] ) ? $settings['wpssg_import_scheduling_run_on'] : '';
			$import_scheduling_run_on       = isset( $_POST['import_scheduling_run_on'] ) ? sanitize_text_field( wp_unslash( $_POST['import_scheduling_run_on'] ) ) : '';
			if ( (string) $wpssg_import_scheduling_run_on !== (string) $import_scheduling_run_on ) {
				self::$instance->wpssg_remove_cron_job( $form_args, 'import' );
			}
			$settings['wpssg_import_scheduling_run_on'] = $import_scheduling_run_on;
			if ( 'recurrence' === (string) $import_scheduling_run_on ) {
				$wpssg_interval                               = array(
					'thirty_minutes' => 'thirty_minutes',
					'once_daily'     => 'daily',
					'twice_daily'    => 'twicedaily',
					'fifteen_days'   => 'fifteen_days',
					'monthly'        => 'monthly',
				);
				$import_schedule_recurrence                   = isset( $_POST['import_schedule_recurrence'] ) ? sanitize_text_field( wp_unslash( $_POST['import_schedule_recurrence'] ) ) : '';
				$wpssg_import_schedule_recurrence             = isset( $settings['wpssg_import_schedule_recurrence'] ) ? $settings['wpssg_import_schedule_recurrence'] : '';
				$settings['wpssg_import_schedule_recurrence'] = $import_schedule_recurrence;
				if ( isset( $wpssg_interval[ $import_schedule_recurrence ] ) ) {
					if ( (string) $import_schedule_recurrence !== (string) $wpssg_import_schedule_recurrence ) {
						self::$instance->wpssg_remove_cron_job( $form_args, 'import' );
					}
					self::$instance->wpssg_set_cron_job( time(), $wpssg_interval[ $import_schedule_recurrence ], $form_args, false, 'import' );
				}
			}
			if ( 'weekly' === (string) $import_scheduling_run_on ) {
				$import_scheduling_weekly_days       = isset( $_POST['import_scheduling_weekly_days'] ) ? sanitize_text_field( wp_unslash( $_POST['import_scheduling_weekly_days'] ) ) : '';
				$wpssg_import_scheduling_weekly_days = isset( $settings['wpssg_import_scheduling_weekly_days'] ) ? $settings['wpssg_import_scheduling_weekly_days'] : '';
				if ( ! empty( $import_scheduling_weekly_days ) ) {
					self::$instance->wpssg_remove_cron_job( $form_args, 'import' );
					$wpssg_weekdays = explode( ',', $import_scheduling_weekly_days );
					if ( is_array( $wpssg_weekdays ) && ! empty( $wpssg_weekdays ) ) {
						$settings['wpssg_import_scheduling_weekly_days'] = $import_scheduling_weekly_days;
						foreach ( $wpssg_weekdays as $weekday ) {
							$startcron = strtotime( 'next ' . $weekday );
							self::$instance->wpssg_set_cron_job( $startcron, 'weekly', $form_args, false, 'import' );
						}
					}
				}
			}
			if ( 'onetime' === (string) $import_scheduling_run_on ) {
				$import_scheduling_date                   = isset( $_POST['import_scheduling_date'] ) ? sanitize_text_field( wp_unslash( $_POST['import_scheduling_date'] ) ) : '';
				$import_scheduling_time                   = isset( $_POST['import_scheduling_time'] ) ? sanitize_text_field( wp_unslash( $_POST['import_scheduling_time'] ) ) : '';
				$wpssg_import_scheduling_date             = isset( $settings['wpssg_import_scheduling_date'] ) ? $settings['wpssg_import_scheduling_date'] : '';
				$wpssg_import_scheduling_time             = isset( $settings['wpssg_import_scheduling_time'] ) ? $settings['wpssg_import_scheduling_time'] : '';
				$settings['wpssg_import_scheduling_date'] = $import_scheduling_date;
				$settings['wpssg_import_scheduling_time'] = $import_scheduling_time;
				if ( ! empty( $import_scheduling_date ) && ! empty( $import_scheduling_time ) ) {
					$starttime = strtotime( $import_scheduling_date . ' ' . $import_scheduling_time );
					if ( ! empty( $wpssg_import_scheduling_date ) && ! empty( $wpssg_import_scheduling_time ) ) {
						$previousstarttime = strtotime( $wpssg_import_scheduling_date . ' ' . $wpssg_scheduling_time );
						if ( (int) $starttime !== (int) $previousstarttime ) {
							self::$instance->wpssg_remove_cron_job( $form_args, 'import' );
						}
					}
					self::$instance->wpssg_set_cron_job( $starttime, '', $form_args, true, 'import' );
				}
			}
		} else {
			self::$instance->wpssg_remove_cron_job( $form_args, 'import' );
		}
		$result = self::$instance->update_feed_meta( $fid, $settings );
		return $settings;
	}
	/**
	 * Set cron job.
	 *
	 * @param int    $startcron .
	 * @param int    $interval .
	 * @param array  $form_args form arguments array including feed id and form id.
	 * @param bool   $once .
	 * @param string $cron_func type of cron funtion need to run.
	 */
	public static function wpssg_set_cron_job( $startcron, $interval, $form_args, $once = false, $cron_func = 'sync' ) {
		$feed_id = isset( $form_args['feed_id'] ) ? $form_args['feed_id'] : '';
		$form_id = isset( $form_args['form_id'] ) ? $form_args['form_id'] : '';
		if ( ! $feed_id || ! $form_id ) {
			return;
		}
		$form_args = array_values( $form_args );
		if ( 'import' === (string) $cron_func ) {
			if ( $once ) {
				wp_schedule_single_event( $startcron, 'wpssg_import_cron_run', $form_args );
			} else {
				wp_schedule_event( $startcron, $interval, 'wpssg_import_cron_run', $form_args );
			}
		} else {
			if ( $once ) {
				wp_schedule_single_event( $startcron, 'wpssg_cron_run', $form_args );
			} else {
				wp_schedule_event( $startcron, $interval, 'wpssg_cron_run', $form_args );
			}
		}
	}
	/**
	 * Remove cron job.
	 *
	 * @param array  $form_args form arguments array including feed id and form id.
	 * @param string $cron_func type of cron funtion need to run.
	 */
	public static function wpssg_remove_cron_job( $form_args, $cron_func = 'sync' ) {
		$feed_id = isset( $form_args['feed_id'] ) ? $form_args['feed_id'] : '';
		$form_id = isset( $form_args['form_id'] ) ? $form_args['form_id'] : '';
		if ( ! $feed_id || ! $form_id ) {
			return;
		}
		$form_args = array_values( $form_args );
		if ( 'import' === (string) $cron_func ) {
			wp_clear_scheduled_hook( 'wpssg_import_cron_run', $form_args );
		} else {
			wp_clear_scheduled_hook( 'wpssg_cron_run', $form_args );
		}
	}

	/** CRON FUNCTION END **/


	/**
	 * Check Mollie plugin is active or not
	 */
	public static function wpssg_is_mollie_active() {
		if ( ! class_exists( 'GF_Mollie' ) ) {
			return false;
		}
		return true;
	}
	/**
	 * Check RedSys For Gravity plugin is active or not
	 */
	public static function wpssg_is_redsysforgravity_active() {
		if ( ! class_exists( 'RedSysForGravity' ) ) {
			return false;
		}
		return true;
	}
	/**
	 * Check GPDFAPI plugin is active or not
	 */
	public static function wpssg_is_gpdfapi_active() {
		if ( ! class_exists( 'GPDFAPI' ) || ! version_compare( get_option( 'gfpdf_current_version' ), '6', '>=' ) ) {
			return false;
		}
		return true;
	}
	/**
	 * Auto sync entries to sheet
	 *
	 * @param int $wpssg_fid feed id.
	 * @param int $form_id form id.
	 */
	public function wpssg_auto_sync_entries( $wpssg_fid, $form_id ) {

		if ( ! $wpssg_fid || ! $form_id ) {
			return;
		}

		$syncformfeed    = RGFormsModel::get_form_meta( $form_id );
		$wpssg_form_data = GFAPI::get_form( $form_id );
		$formfeed        = self::$instance->get_feeds( $form_id );
		$form_field_type = self::$instance->wpssg_form_field_type( $syncformfeed );

		foreach ( $formfeed as $feedsetting ) {
			if ( (int) $feedsetting['id'] === (int) $wpssg_fid ) {
				$settings = $feedsetting['meta'];
			}
		}

		$spreadsheetid = isset( $settings['spreedsheets'] ) ? $settings['spreedsheets'] : '';
		$range         = isset( $settings['sheetselect'] ) ? $settings['sheetselect'] : '';

		$sheet_data = self::$instance->wpssg_check_sheet_exist( $spreadsheetid, $range );
		if ( '' === (string) $sheet_data['spreadsheetid'] || '' === (string) $sheet_data['sheetname'] ) {
			return;
		}

		$paging = array(
			'offset'    => 0,
			'page_size' => 999999,
		);

		$search_criteria = array();

		// default.
		$sorting = array( 'direction' => 'ASC' );

		$entry = GFAPI::get_entries( $form_id, $search_criteria, $sorting, $paging );

		$entryids = array_column( $entry, 'id' );

		$formmapfieldsval      = self::$instance->wpssg_get_mapfields( $spreadsheetid, $range );
		$allentry              = self::$instance_api->get_row_list( $spreadsheetid, $range . '!A:Z' );
		$data                  = $allentry->getValues();
		$data                  = array_map(
			function( $element ) {
				if ( isset( $element['0'] ) ) {
					return $element['0'];
				} else {
					return '';
				}
			},
			$data
		);
		$data                  = self::$instance->wpssg_convert_int( $data );
		$newenrty              = array();
		$is_condition_enabled  = rgar( $settings, 'feed_condition_conditional_logic' ) === '1';
		$is_creditcard_enabled = self::$instance->wpssg_get_credit_card_field( $wpssg_form_data );
		$logic                 = rgars( $settings, 'feed_condition_conditional_logic_object/conditionalLogic' );

		foreach ( $entryids as $entryid ) {
			if ( in_array( (int) $entryid, $data, true ) ) {
				continue;
			} else {
				foreach ( $entry as $addentry ) {
					if ( (int) $entryid === (int) $addentry['id'] ) {
						if ( $is_condition_enabled && ! empty( $logic ) ) {
							$check = GFCommon::evaluate_conditional_logic( $logic, $wpssg_form_data, $addentry );
							if ( ! $check ) {
								continue;
							}
						}
						if ( $is_creditcard_enabled ) {
							if ( isset( $settings['paymentstatus'] ) && $settings['paymentstatus'] ) {
								if ( ! in_array( $addentry['payment_status'], array( 'Paid', 'Approved' ), true ) ) {
									continue;
								}
							}
						}
						$newenrty[] = $addentry;
					}
				}
			}
		}
		$total_entries = count( $newenrty );
		$values_data   = array();

		$val = self::$instance->wpssg_prepared_headers( $formmapfieldsval, $settings );

		$response            = self::$instance_api->get_sheet_listing( $spreadsheetid );
		$existingsheetsnames = self::$instance_api->get_sheet_list( $response );

		$sheetid = $existingsheetsnames[ $range ];

		$startindex = count( $data ) + 1;
		$endindex   = $startindex + $total_entries;

		$param               = array();
		$param['sheetid']    = $sheetid;
		$param['startindex'] = $startindex;
		$param['endindex']   = $endindex;

		$batchupdaterequest = self::$instance_api->insertdimensionobject( $param );

		$requestobject                  = array();
		$requestobject['spreadsheetid'] = $spreadsheetid;
		$requestobject['requestbody']   = $batchupdaterequest;
		$response                       = self::$instance_api->formatsheet( $requestobject );

		$entry_limit = 500;
		if ( (int) $total_entries < 2000 ) {
			if ( (int) $total_entries > 50 ) {
				$entry_limit = 50;
			}
			if ( (int) $total_entries <= 50 ) {
				$entry_limit = (int) $total_entries;
			}
		}
		$j           = 0;
		$entry_count = $entry_limit;
		$total_limit = $entry_limit;
		try {
			for ( $k = 0;$k < (float) ( $total_entries / $entry_limit );$k++ ) {
				set_time_limit( 999 );
				$values_data = array();
				for ( $i = $j;$i < ( $total_limit );$i++ ) {
					$values_data[] = self::$instance->wpssg_prepared_values( $wpssg_form_data, $val, $newenrty[ $i ], $settings );
				}
				if ( is_array( $values_data ) && ! empty( $values_data ) ) {
					try {
						/*Append entry*/
						$rangetoupdate = $range . '!A' . $startindex;

						$requestbody = self::$instance_api->valuerangeobject( $values_data );
						$params      = self::$instance->get_row_format( $settings );
						$param       = self::$instance_api->setparamater( $spreadsheetid, $rangetoupdate, $requestbody, $params );
						$response    = self::$instance_api->updateentry( $param );
					} catch ( Exception $e ) {
						return;
					}
				}
				$j          += $entry_limit;
				$startindex += $entry_limit;
				$total_limit = ( $entry_count + $entry_limit ) < $total_entries ? ( $entry_count + $entry_limit ) : $total_entries;
				$entry_count = ( $entry_count + $entry_limit ) < $total_entries ? ( $entry_count + $entry_limit ) : ( $total_entries - $entry_limit );
			}
		} catch ( Exception $e ) {
			return;
		}
	}
}
