<?php
	/**
	 * Main WPSyncSheets_Gravity_Forms namespace.
	 *
	 * @since 1.0.0
	 * @package wpsyncsheets-for-gravityforms
	 */

namespace WPSyncSheets_Gravity_Forms {
	/**
	 * Main WPSyncSheets_Gravity_Forms class.
	 *
	 * @since 1.0.0
	 * @package wpsyncsheets-for-gravityforms
	 */
	final class WPSyncSheets_Gravity_Forms {
		/**
		 * Store Instance of class.
		 *
		 * @since 1.0.0
		 *
		 * @var \WPSyncSheets_Gravity_Forms\WPSyncSheets_Gravity_Forms
		 */
		private static $instance;

		/**
		 * Plugin version for enqueueing, etc.
		 * The value is got from WPSSG_VERSION constant.
		 *
		 * @since 1.0.0
		 *
		 * @var string
		 */
		public $version = '';

		/**
		 * Main WPSyncSheets_Gravity_Forms Instance.
		 *
		 * Only one instance of WPSyncSheets_Gravity_Forms exists in memory at any one time.
		 * Also prevent the need to define globals all over the place.
		 *
		 * @since 1.0.0
		 *
		 * @return WPSyncSheets_Gravity_Forms
		 */
		public static function instance() {

			if ( null === self::$instance || ! self::$instance instanceof self ) {

				self::$instance = new self();
				self::$instance->constants();
				self::$instance->includes();

				add_action( 'init', array( self::$instance, 'load_textdomain' ), 10 );
			}

			return self::$instance;
		}

		/**
		 * Setup plugin constants.
		 * All the path/URL related constants are defined in main plugin file.
		 *
		 * @since 1.0.0
		 */
		private function constants() {
			$this->version = WPSSG_VERSION;
		}

		/**
		 * Load the plugin language files.
		 *
		 * @since 1.0.0
		 */
		public function load_textdomain() {

			// If the user is logged in, unset the current text-domains before loading our text domain.
			// This feels hacky, but this way a user's set language in their profile will be used,
			// rather than the site-specific language.
			if ( is_user_logged_in() ) {
				unload_textdomain( 'wpssg' );
			}
			load_plugin_textdomain( 'wpssg', false, WPSSG_DIRECTORY . '/assets/languages/' );
		}

		/**
		 * Include files.
		 *
		 * @since 1.0.0
		 */
		private function includes() {

			// Global Includes.

			require_once WPSSG_PATH . '/includes/class-wpssg-google-api.php';
			require_once WPSSG_PATH . '/includes/class-wpssg-feed-settings.php';
			require_once WPSSG_PATH . '/includes/class-wpssg-google-api-functions.php';
			require_once WPSSG_PATH . '/includes/class-wpssg-plugin-settings.php';
			require_once WPSSG_PATH . '/includes/class-wpssg-plugin-update.php';
			require_once WPSSG_PATH . '/includes/class-wpssg-licence.php';
		}
	}
}

namespace {

	/**
	 * The function which returns the one WPSSG instance.
	 *
	 * @since 1.0.0
	 *
	 * @return WPSSG\wpssg
	 */
	function wpssg() {
		return WPSyncSheets_Gravity_Forms\WPSyncSheets_Gravity_Forms::instance();
	}
	class_alias( 'WPSyncSheets_Gravity_Forms\WPSyncSheets_Gravity_Forms', 'WPSyncSheets_Gravity_Forms' );
}
